import math
import contextlib
import io
import os
import numpy as np

from PIL import Image
from matplotlib import pyplot as plt
from tabulate import tabulate

from .model import Model
from .simulator import Simulator
from .gspread_client import GspreadClient
from .gdrive_client import GdriveClient
from .utils import report_service_conf, set_report_service_conf, get_grandparent_folder_path

"""
	This module provides functions and classes that demonstrate the functionality of the slacgs package. 
	
	Global Variables:
		SCENARIO1 (list): list of parameters for scenario 1
		SCENARIO2 (list): list of parameters for scenario 2
		SCENARIO3 (list): list of parameters for scenario 3
		SCENARIO4 (list): list of parameters for scenario 4
		SCENARIOS (list): list of scenarios for simulation
		GDC (GdriveClient): Google Drive Client for reports service
	
	Functions:
		start_google_drive_service(): starts Google Drive service
		run_experiment_simulation( scenario, model, simulator, gdc ): runs experiment simulation
				
	
"""

## define list of parameters for scenario 1
SCENARIO1 = [[1, 1, round(1 + 0.1 * sigma3, 2), 0, 0, 0] for sigma3 in range(3, 10)]
SCENARIO1 += [[1, 1, sigma3 / 2, 0, 0, 0] for sigma3 in range(4, 11, 1)]
SCENARIO1 += [[1, 1, sigma3, 0, 0, 0] for sigma3 in range(6, 14, 1)]

## define list of parameters for scenario 2
SCENARIO2 = [[1, 1, 2, round(rho12 * 0.1, 1), 0, 0] for rho12 in range(-8, 9)]

## define list of parameters for scenario 3
RHO_12 = 0
SCENARIO3 = []
for r in range(-8, 8):
	if abs(round(0.1 * r, 1)) < math.sqrt((1 + RHO_12) / 2):
		SCENARIO3 += [[1, 1, 2, RHO_12, round(0.1 * r, 1), round(0.1 * r, 1)]]

## define list of parameters for scenario 4
RHO_12 = -0.1
SCENARIO4 = []
for r in range(-8, 8):
	if abs(round(0.1 * r, 1)) < math.sqrt((1 + RHO_12) / 2):
		SCENARIO4 += [[1, 1, 2, RHO_12, round(0.1 * r, 1), round(0.1 * r, 1)]]

## create list of scenarios
SCENARIOS = [SCENARIO1, SCENARIO2, SCENARIO3, SCENARIO4]

## free global variables
r = RHO_12 = None

# initialize Google Drive Client global variable for demo reports service
GDC = None


def start_google_drive_service(password=None, user_email=None):
	"""start Google Drive Client for demo report service

	Parameters:
		password (str): password for slacgs report service
		user_email (str): email for Google Drive account to be used for report service

	Notes:
		if password and user_email are None, the report_service_conf dictionary will be used to get the password and user_email
	"""

	## create GdriveClient object and connect to Google Drive for reports service
	if report_service_conf['drive_service'] is None:
		set_report_service_conf(user_google_account_email=user_email, slacgs_password=password)

	global GDC
	if GDC is None:
		GDC = GdriveClient(report_service_conf['drive_service'], report_service_conf['spreadsheet_service'],
		                   report_service_conf['user_email'])

	if GDC.gdrive_account_email:
		if not GDC.folder_exists('slacgs.demo.' + GDC.gdrive_account_email):
			folder_id = GDC.create_folder('slacgs.demo.' + GDC.gdrive_account_email)
		else:
			folder_id = GDC.get_folder_id_by_name('slacgs.demo.' + GDC.gdrive_account_email)

	GDC.share_folder_with_gdrive_account(folder_id)


def run_experiment_simulation(start_scenario=1, verbose=True):
	""" run a simulation for one of the experiment scenarios and return True if there are still parameters to be simulated and False otherwise.

	- Reports with results will be stored in a Google Spreadsheet for each:  Experiment Scenario, Custom Experiment Scenario	and another one for the Custom Simulations.
	- The Spreadsheets are stored in a Google Drive folder named 'slacgs.demo.<user_email>'	owned by slacgs' google service	account and shared with the user's Google Drive account.
	- Also, images with data visualization will be exported to a local folder inside user's local folder (<user>/slacgs/images/ or /content/slacgs/images (for G-colab) )

	Reports Exported:
		- Loss Report: Contains mainly results focused on Loss Functions evaluations for each dimensionality of the model.
		- Compare Resport: Contains mainly results focused on comparing the performance of the Model using 2 features and 3 features.
		- Home Report (Scenario): Contains results from all simulations in a Scenario and links to the other reports. (available only for comparison between 2D and 3D)

	Images Exported (<user>/slacgs/images/ or /content/slacgs/images [for G-colab] ):
		- Scenario Data plots .gif: Contains a gif with all plots with the data points (n = 1024, dims=[2,3] ) generated for all Models in an Experiment Scenario.
		- Simulation Data plot .png: Contains a plot with the data points (n = 1024, dims=[2,3] ) generated for a Model in a Simulation.
		- Simulation Loss plot .png: Contains a plot with the loss values (Theoretical, Empirical with Train Data, Empirical with Test data) generated for a Model in a Simulation.

	Loss Functions:
		- Theoretical Loss: estimated using probability theory
		- Empirical Loss with Train Data: estimated using empirical approach with train data
		- Empirical Loss with Test Data: estimated using empirical approach with test data

	Dimensions simulated:
		- 1D: 1 feature
		- 2D: 2 features
		- 3D: 3 features

	Dimensions compared:
		- 2D vs 3D: 2 features vs 3 features

	Cardinalities simulated:
		- N = [2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]  (+ [2048, 4096, 8192] if L_3 > L_2)

	Parameters:
		start_scenario (int): scenario to start the simulation (default is 1)
		verbose (bool): if True, print simulation progress to stdout (default is True)

	Returns:
		bool: True if there are still parameters to be simulated and False otherwise

	Raises:
		ValueError: if start_scenario is not between 1 and 4
		TypeError: if start_scenario is not an int


	Example:
		>>> from slacgs.demo import *

		>>> # opt-1. set report service configuration with your own google cloud service account key file
		>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
		>>> set_report_service_conf(PATH_TO_KEY_FILE)

		>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
		>>> set_report_service_conf()

		>>> run_experiment_simulation()

	"""

	if not isinstance(start_scenario, int):
		raise TypeError("start_scenario must be an int")

	if not 1 <= start_scenario <= 4:
		raise ValueError("start_scenario must be between 1 and 4")

	start_google_drive_service()

	## define folder name for storing reports
	REPORT_FOLDER_NAME = 'slacgs.demo.' + GDC.gdrive_account_email

	## create folder if it doesn't exist
	if not GDC.folder_exists(REPORT_FOLDER_NAME):
		folder_id = GDC.create_folder(REPORT_FOLDER_NAME)  # create folder
		GDC.share_folder_with_gdrive_account(folder_id, verbose=verbose)  # share folder with user's google drive account

	SPREADSHEET_TITLE = 'scenario1'
	## create spreadsheet for the first simulation if it doesn't exist
	if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
		spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE, verbose=verbose)
		folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
		GDC.move_file_to_folder(spreadsheet_id, folder_id, verbose=verbose)
		gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)
		PARAM = SCENARIOS[0][0]
	else:  # if spreadsheet already exists, then find the first parameter that is not in the spreadsheet report home
		for i in range(start_scenario - 1, len(SCENARIOS)):
			SPREADSHEET_TITLE = 'scenario' + str(i + 1)

			## create spreadsheet if it doesn't exist
			if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
				spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE, verbose=verbose)
				folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
				GDC.move_file_to_folder(spreadsheet_id, folder_id, verbose=verbose)

			gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)

			## retrieve the first parameter that is not in the spreadsheet report home
			PARAM = None
			for j in range(len(SCENARIOS[i])):
				params = SCENARIOS[i][j]
				if gsc.param_not_in_home(params):
					PARAM = params
					break

			## if all parameters are in the spreadsheet report home, then go to the next spreadsheet
			if PARAM:
				break

	if not PARAM:
		print("All parameters have been simulated. Please check your google drive section: 'Shared with me' for results.")
		return False

	## create model object
	model = Model(PARAM)

	## create simulator object
	slacgs = Simulator(model, verbose=verbose)

	## run simulation
	slacgs.run()

	## upload png images to drive
	slacgs.report.upload_loss_plot_to_drive(GDC, verbose=verbose)
	slacgs.model.upload_data_points_plot_to_google_drive(GDC, verbose=verbose)

	## write results to spreadsheet
	slacgs.report.write_to_spreadsheet(gsc, verbose=verbose)

	return True


def add_simulation_to_experiment_scenario_spreadsheet(params, scenario_number, verbose=True):
	""" add simulation results to one of the experiment scenario spreadsheets

	Parameters:
		params (list[float|int] or tuple[float|int]): a list containnong Sigmas and Rhos
		scenario_number (int): scenario number
		verbose (bool): if True, print simulation progress (default: True)

	Returns:
		None

	Raises:
		TypeError:
			if scenario is not an int;
			if params is not a list[int|float] or tuple[int|float]
		ValueError:
			if params is not a valid parameter list;
			if scenario is not between 1 and 4

	See Also:
		:func:`slacgs.demo.run_experiment_simulation`

	Example:
		>>> from slacgs.demo import *

		>>> # opt-1. set report service configuration with your own google cloud service account key file
		>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
		>>> set_report_service_conf(PATH_TO_KEY_FILE)

		>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
		>>> set_report_service_conf()

		>>> scenario_number = 1
		>>> params = [1, 1, 2.1, 0, 0, 0]
		>>> add_simulation_to_experiment_scenario_spreadsheet(params, scenario_number)

		>>> scenario_number = 2
		>>> params = [1, 1, 2, -0.15, 0, 0]
		>>> add_simulation_to_experiment_scenario_spreadsheet(params, scenario_number)

		>>> scenario_number = 3
		>>> params = [1, 1, 2, 0, 0.15, 0.15]
		>>> add_simulation_to_experiment_scenario_spreadsheet(params, scenario_number)

		>>> scenario_number = 4
		>>> params = [1, 1, 2, -0.1, 0.15, 0.15]
		>>> add_simulation_to_experiment_scenario_spreadsheet(params, scenario_number)


	"""

	if not isinstance(scenario_number, int):
		raise TypeError("scenario must be an int")

	if not isinstance(params, (list, tuple)):
		raise TypeError("params must be a list or tuple")

	if not all(isinstance(x, (int, float)) for x in params):
		raise TypeError("params must be a list or tuple of int or float")

	if scenario_number < 1 or scenario_number > 4:
		raise ValueError("scenario must be between 1 and 4")

	if len(params) != 6:
		raise ValueError("params must be a list or tuple of 6 elements for this experiment")

	if scenario_number == 1:
		if params[0] != 1 or params[1] != 1 or params[3] != 0 or params[4] != 0 or params[5] != 0:
			raise ValueError(
				"for scenario 1, params must be a list or tuple of 6 elements where params[0] = 1, params[1] = 1, params[3] = 0, params[4] = 0 and params[5] = 0")

	elif scenario_number == 2:
		if params[0] != 1 or params[1] != 1 or params[2] != 2 or params[4] != 0 or params[5] != 0:
			raise ValueError(
				"for scenario 2, params must be a list or tuple of 6 elements where params[0] = 1, params[1] = 1, params[2] = 2, params[4] = 0 and params[5] = 0")
		if params[3] < -0.8 or params[3] > 0.8:
			raise ValueError("for scenario 2, params[3] must be between -0.8 and 0.8")

	elif scenario_number == 3:
		if params[0] != 1 or params[1] != 1 or params[2] != 2 or params[3] != 0:
			raise ValueError(
				"for scenario 3, params must be a list or tuple of 6 elements where params[0] = 1, params[1] = 1, params[2] = 2 and params[3] = 0")
		if params[4] < -0.7 or params[4] > 0.7 or params[4] != params[5]:
			raise ValueError(
				"for scenario 3, params[4] must be between -0.7 and 0.7 and params[4] must be equal to params[5]")

	elif scenario_number == 4:
		if params[0] != 1 or params[1] != 1 or params[2] != 2 or params[3] != -0.1:
			raise ValueError(
				"for scenario 4, params must be a list or tuple of 6 elements where params[0] = 1, params[1] = 1, params[2] = 2 and params[3] = -0.1")
		if params[4] < -0.6 or params[4] > 0.6 or params[4] != params[5]:
			raise ValueError(
				"for scenario 4, params[4] must be between -0.6 and 0.6 and params[4] must be equal to params[5]")

	## update scenario gif
	save_scenario_figures_as_gif([params], scenario_number, verbose=verbose)

	start_google_drive_service()

	## define folder name for storing reports
	REPORT_FOLDER_NAME = 'slacgs.demo.' + GDC.gdrive_account_email

	## create folder if it doesn't exist
	if not GDC.folder_exists(REPORT_FOLDER_NAME):
		folder_id = GDC.create_folder(REPORT_FOLDER_NAME, verbose=verbose)  # create folder
		GDC.share_folder_with_gdrive_account(folder_id, verbose=verbose)  # share folder with user's google drive account

	## define spreadsheet title
	SPREADSHEET_TITLE = 'scenario' + str(scenario_number)

	## create spreadsheet if it doesn't exist
	if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
		spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE, verbose=verbose)
		folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
		GDC.move_file_to_folder(spreadsheet_id, folder_id, verbose=verbose)

	## create gspread client object
	gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)

	## create model object
	model = Model(params)

	## create simulator object
	slacgs = Simulator(model, verbose=verbose)

	## run simulation
	slacgs.run()

	## upload png images to drive
	slacgs.report.upload_loss_plot_to_drive(GDC, verbose=verbose)
	slacgs.model.upload_data_points_plot_to_google_drive(GDC, verbose=verbose)

	## write results to spreadsheet
	slacgs.report.write_to_spreadsheet(gsc, verbose=verbose)


def run_custom_scenario(scenario_list, scenario_number, dims_to_simulate=None, dims_to_compare=None, verbose=True):
	""" run a custom scenario and write the results to a Google Spreadsheet shared with the user.	A Scenario is a list with params to simulate.

	- Reports with results will be stored in a Google Spreadsheet for each:  Experiment Scenario, Custom Experiment Scenario	and another one for the Custom Simulations.
	- The Spreadsheets are stored in a Google Drive folder named 'slacgs.demo.<user_email>'	owned by slacgs' google service	account and shared with the user's Google Drive account.
	- Also, images with data visualization will be exported to a local folder inside user's local folder (<user>/slacgs/images/ or /content/slacgs/images (for G-colab) )

	Reports Exported:
		- Loss Report: Contains mainly results focused on Loss Functions evaluations for each dimensionality of the model.
		- Compare Resport: Contains mainly results focused on comparing the performance of the Model using 2 features and 3 features.
		- Home Report (Scenario): Contains results from all simulations in a Scenario and links to the other reports. (available only for comparison between 2D and 3D)

	Images Exported (<user>/slacgs/images/ or /content/slacgs/images [for G-colab] ):
		- Scenario Data plots .gif: Contains a gif with all plots with the data points (n = 1024, dims=[2,3] ) generated for all Models in an Experiment Scenario.
		- Simulation Data plot .png: Contains a plot with the data points (n = 1024, dims=[2,3] ) generated for a Model in a Simulation.
		- Simulation Loss plot .png: Contains a plot with the loss values (Theoretical, Empirical with Train Data, Empirical with Test data) generated for a Model in a Simulation.

	Loss Functions:
		- Theoretical Loss: estimated using probability theory
		- Empirical Loss with Train Data: estimated using empirical approach with train data
		- Empirical Loss with Test Data: estimated using empirical approach with test data

	Cardinalities simulated:
		- N = [2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]  (+ [2048, 4096, 8192] if L_3 > L_2)

	Parameters:
		scenario_list (list[list[float|int]] or tuple[list[float|int]]): scenario list
		scenario_number (int): scenario number
		dims_to_simulate (list[int] or tuple[int]): dimensionalities to simulate (if None, all possible dimensionalities will be simulated) (default: None)
		dims_to_compare (list[int] or tuple[int]): dimensionalities to compare (if None, the last two dimensionalities of dims_to_simulate will be compared) (default: None)
		verbose (bool): whether to print messages to stdout or not

	Returns:
		bool: True if simulation was successful

	Raises:
		TypeError:
			if scenario_list is not a list[list[float|int]] or tuple[list[float|int]];
			if scenario_number is not an int;
			if dims_to_simulate is not a list[int] or tuple[int];
			if dims_to_compare is not a list[int] or tuple[int];

		ValueError:
			if scenario_number is not a valid scenario number;
			if dims_to_compare is not a subset of dims_to_simulate;
			if dims_to_compare length is not 2;


	Example:
		>>> from slacgs.demo import *

		>>> # opt-1. set report service configuration with your own google cloud service account key file
		>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
		>>> set_report_service_conf(PATH_TO_KEY_FILE)

		>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
		>>> set_report_service_conf()

		>>> scenario_list = [[1,1,3,round(0.1*rho,1),0,0] for rho in range(-1,2)]
		>>> scenario_number = 5
		>>> run_custom_scenario(scenario_list, scenario_number)
	"""

	if not isinstance(scenario_list, (list, tuple)):
		raise TypeError("scenario must be a list or tuple")

	if not all(isinstance(x, (list, tuple)) for x in scenario_list):
		raise TypeError("scenario must be a list or tuple of list or tuple")

	if not all(isinstance(x, (int, float)) for y in scenario_list for x in y):
		raise TypeError("scenario must be a list or tuple of list or tuple of int or float")

	if not isinstance(scenario_number, int):
		raise TypeError("scenario_number must be an int")

	if dims_to_simulate and not isinstance(dims_to_simulate, (list, tuple)):
		raise TypeError("dims_to_simulate must be a list or tuple")

	if dims_to_simulate and not all(isinstance(x, int) for x in dims_to_simulate):
		raise TypeError("dims_to_simulate must be a list or tuple of int")

	if dims_to_compare and not isinstance(dims_to_compare, (list, tuple)):
		raise TypeError("dims_to_compare must be a list or tuple")

	if dims_to_compare and not all(isinstance(x, int) for x in dims_to_compare):
		raise TypeError("dims_to_compare must be a list or tuple of int")

	if scenario_number < 5:
		raise ValueError("Custom scenario_number must be >= 5")

	start_google_drive_service()

	## create Model objects to test each parameter set before continuing
	models = []
	for params in scenario_list:
		models.append(Model(params))

	## create Simulator objects to test each parameter set before continuing
	simulators = []
	for model in models:
		simulators.append(Simulator(model, dims=dims_to_simulate, dims_to_compare=dims_to_compare, verbose=verbose))

	save_scenario_figures_as_gif(scenario_list, scenario_number, verbose=verbose)

	if dims_to_compare and not all(dim in dims_to_simulate for dim in dims_to_compare):
		raise ValueError("dims_to_compare must be a subset of dims_to_simulate")

	if dims_to_compare and len(dims_to_compare) != 2:
		raise ValueError("dims_to_compare length must be 2")

	## define folder name for storing reports
	REPORT_FOLDER_NAME = 'slacgs.demo.' + GDC.gdrive_account_email

	## create folder if it doesn't exist
	if not GDC.folder_exists(REPORT_FOLDER_NAME):
		folder_id = GDC.create_folder(REPORT_FOLDER_NAME, verbose=verbose)  # create folder
		GDC.share_folder_with_gdrive_account(folder_id, verbose=verbose)  # share folder with user's google drive account

	## define spreadsheet title
	SPREADSHEET_TITLE = 'custom_scenario' + str(scenario_number)

	## create spreadsheet if it doesn't exist
	if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
		spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE, verbose=verbose)
		folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
		GDC.move_file_to_folder(spreadsheet_id, folder_id, verbose=verbose)

	## create gspread client object
	gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)

	for slacgs in simulators:
		if dims_to_compare == (2, 3) or dims_to_compare == [2, 3]:
			if gsc.param_not_in_home(slacgs.model.params):
				## run simulation
				slacgs.run()

				## upload png images to drive
				slacgs.report.upload_loss_plot_to_drive(GDC, verbose=verbose)
				slacgs.model.upload_data_points_plot_to_google_drive(GDC, verbose=verbose)

				## write results to spreadsheet
				slacgs.report.write_to_spreadsheet(gsc, verbose=verbose)
			else:
				continue

		else:
			## run simulation
			slacgs.run()

			## upload png images to drive
			slacgs.report.upload_loss_plot_to_drive(GDC, verbose=verbose)
			slacgs.model.upload_data_points_plot_to_google_drive(GDC, verbose=verbose)

			## write results to spreadsheet
			slacgs.report.write_to_spreadsheet(gsc, verbose=verbose)

	return True


def add_simulation_to_custom_scenario_spreadsheet(params, scenario_number, dims_to_simulate=None, dims_to_compare=None,
                                                  verbose=True):
	""" add a simulation to a custom  scenario spreadsheet

	Parameters:
		params (list[float|int] or tuple[float|int]): a list containnong Sigmas and Rhos
		scenario_number (int): scenario number
		dims_to_simulate (list[int] or tuple[int]): dimensionalities to simulate (if None, all dimensionalities will be simulated) (default: None)
		dims_to_compare (list[int] or tuple[int]): dimensionalities to compare (if None, the last two dimensionalities of dims_to_simulate will be compared) (default: None)
		verbose (bool): whether to print progress to console output (default: True)

	Returns:
		bool: True if successful, False otherwise

	Raises:
		TypeError:
			if params is not a list[float|int] or tuple[float|int];
			if scenario_number is not an int;
			if dims_to_simulate is not a list[int] or tuple[int];
			if dims_to_compare is not a list[int] or tuple[int]

		ValueError:
			if scenario_number is < 5;
			if dims_to_compare is not a subset of dims_to_simulate;
			if dims_to_compare length is not 2


	Example:

		>>> from slacgs.demo import *

		>>> # opt-1. set report service configuration with your own google cloud service account key file
		>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
		>>> set_report_service_conf(PATH_TO_KEY_FILE)

		>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
		>>> set_report_service_conf()

		>>> params = (1, 1, 3, -0.2, 0, 0)
		>>> scenario_number = 5
		>>> add_simulation_to_custom_scenario_spreadsheet(params, scenario_number)


	"""

	if not isinstance(params, (list, tuple)):
		raise TypeError("params must be a list or tuple")

	if not isinstance(scenario_number, int):
		raise TypeError("scenario_number must be an int")

	if dims_to_simulate and not isinstance(dims_to_simulate, (list, tuple)):
		raise TypeError("dims_to_simulate must be a list or tuple")

	if dims_to_compare and not isinstance(dims_to_compare, (list, tuple)):
		raise TypeError("dims_to_compare must be a list or tuple")

	if scenario_number < 5:
		raise ValueError("scenario_number must be >= 5 for a Custom Scenario")

	if dims_to_compare and dims_to_simulate and not set(dims_to_compare).issubset(set(dims_to_simulate)):
		raise ValueError("dims_to_compare must be a subset of dims_to_simulate")

	if dims_to_compare and len(dims_to_compare) != 2:
		raise ValueError("dims_to_compare length must be 2")

	## update scenario gif
	save_scenario_figures_as_gif([params], scenario_number, verbose=verbose)

	start_google_drive_service()

	## create Model object to test parameter set before continuing
	model = Model(params)

	## create Simulator object to test parameters before continuing
	slacgs = Simulator(model, dims=dims_to_simulate, dims_to_compare=dims_to_compare, verbose=verbose)

	## define folder name for storing reports
	REPORT_FOLDER_NAME = 'slacgs.demo.' + GDC.gdrive_account_email

	## create folder if it doesn't exist
	if not GDC.folder_exists(REPORT_FOLDER_NAME):
		folder_id = GDC.create_folder(REPORT_FOLDER_NAME, verbose=verbose)  # create folder
		GDC.share_folder_with_gdrive_account(folder_id, verbose=verbose)  # share folder with user's google drive account

	## define spreadsheet title
	SPREADSHEET_TITLE = 'custom_scenario' + str(scenario_number)

	## create spreadsheet if it doesn't exist
	if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
		spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE)
		folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
		GDC.move_file_to_folder(spreadsheet_id, folder_id, verbose=verbose)

	## create gspread client object
	gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)

	## run simulation
	slacgs.run()

	## upload png images to drive
	slacgs.report.upload_loss_plot_to_drive(GDC, verbose=verbose)
	slacgs.model.upload_data_points_plot_to_google_drive(GDC, verbose=verbose)

	## write results to spreadsheet
	slacgs.report.write_to_spreadsheet(gsc, verbose=verbose)

	return True


def run_custom_simulation(params, dims_to_simulate=None, dims_to_compare=None, verbose=True):
	""" run a custom simulation for any dimensionality and cardinality

	- Reports with results will be stored in a Google Spreadsheet for each:  Experiment Scenario, Custom Experiment Scenario	and another one for the Custom Simulations.
	- The Spreadsheets are stored in a Google Drive folder named 'slacgs.demo.<user_email>'	owned by slacgs' google service	account and shared with the user's Google Drive account.
	- Also, images with data visualization will be exported to a local folder inside user's local folder (<user>/slacgs/images/ or /content/slacgs/images (for G-colab) )

	Reports Exported:
		- Loss Report: Contains mainly results focused on Loss Functions evaluations for each dimensionality of the model.
		- Compare Resport: Contains mainly results focused on comparing the performance of the Model using 2 features and 3 features.
		- Home Report (Scenario): Contains results from all simulations in a Scenario and links to the other reports. (available only for comparison between 2D and 3D)

	Images Exported (<user>/slacgs/images/ or /content/slacgs/images [for G-colab] ):
		- Scenario Data plots .gif: Contains a gif with all plots with the data points (n = 1024, dims=[2,3] ) generated for all Models in an Experiment Scenario.
		- Simulation Data plot .png: Contains a plot with the data points (n = 1024, dims=[2,3] ) generated for a Model in a Simulation.
		- Simulation Loss plot .png: Contains a plot with the loss values (Theoretical, Empirical with Train Data, Empirical with Test data) generated for a Model in a Simulation.

	Loss Functions:
		- Theoretical Loss: estimated using probability theory
		- Empirical Loss with Train Data: estimated using empirical approach with train data
		- Empirical Loss with Test Data: estimated using empirical approach with test data

	Cardinalities simulated:
		- N = [2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]  (+ [2048, 4096, 8192] if L_3 > L_2)

	Parameters:
		params (list[float|int] or tuple[float|int]): a list containnong Sigmas and Rhos
		dims_to_simulate (list[int] or tuple[int]): dimensionalities to simulate (if None, all dimensionalities will be simulated) (default: None)
		dims_to_compare (list[int] or tuple[int]): dimensionalities to compare (if None, the last two dimensionalities of dims_to_simulate will be compared) (default: None)
		verbose (bool): whether to print progress to console output (default: True)

	Returns:
		bool: True if successful, False otherwise

	Raises:
		TypeError:
			if params is not a list[float|int] or tuple[float|int];
			if dims_to_compare is not a list[int] or tuple[int]

		ValueError:
			if dims_to_compare length is not 2;
			if dims_to_compare is not a subset of dims_to_simulate


	Examples:
		>>> from slacgs.demo import *

		>>> # opt-1. set report service configuration with your own google cloud service account key file
		>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
		>>> set_report_service_conf(PATH_TO_KEY_FILE)

		>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
		>>> set_report_service_conf()

		>>> ## 2 features
		>>> params = [1, 2, 0.4]
		>>> run_custom_simulation(params)

		>>> ## 3 features
		>>> params = [1, 1, 4, -0.2, 0.1, 0.1]
		>>> run_custom_simulation(params)

		>>> ## 4 features
		>>> params = [1, 1, 1, 2, 0, 0, 0, 0, 0, 0]
		>>> run_custom_simulation(params)

		>>> ## 5 features
		>>> params = [1, 1, 2, 2, 2, -0.3, -0.2, -0.1, 0, 0.1, 0.2, 0.2, 0, 0, 0]
		>>> dims_to_compare = (2, 5)
		>>> run_custom_simulation(params, dims_to_compare)

		>>> ## 6 features
		>>> params = [1, 2, 3, 4, 5, 6, -0.3, -0.3, -0.2, -0.2, -0.1, -0.1, 0, 0, 0.1, 0.1, 0.2, 0.2, 0.3, 0.3, 0.4]
		>>> run_custom_simulation(params)

	"""

	if not isinstance(params, (list, tuple)):
		raise TypeError("params must be a list or tuple")

	if not all(isinstance(x, (int, float)) for x in params):
		raise TypeError("params must be a list or tuple of int or float")

	if dims_to_compare and not isinstance(dims_to_compare, (list, tuple)):
		raise TypeError("dims_to_compare must be a list or tuple")

	if dims_to_compare and not all(isinstance(x, int) for x in dims_to_compare):
		raise TypeError("dims_to_compare must be a list or tuple of int")

	if dims_to_compare and len(dims_to_compare) != 2:
		raise ValueError("dims_to_compare must be a list or tuple of length 2")

	if dims_to_compare and dims_to_simulate and not set(dims_to_compare).issubset(set(dims_to_simulate)):
		raise ValueError("dims_to_compare must be a subset of dims_to_simulate")

	## initialize gdrive client if it hasn't been initialized yet
	start_google_drive_service()

	## create model object
	model = Model(params)

	## create simulator object
	slacgs = Simulator(model, dims=dims_to_simulate, dims_to_compare=dims_to_compare, verbose=verbose)

	## define folder name for storing reports
	REPORT_FOLDER_NAME = 'slacgs.demo.' + GDC.gdrive_account_email

	## create folder if it doesn't exist
	if not GDC.folder_exists(REPORT_FOLDER_NAME):
		folder_id = GDC.create_folder(REPORT_FOLDER_NAME, verbose=verbose)  # create folder
		GDC.share_folder_with_gdrive_account(folder_id, verbose=verbose)  # share folder with user's google drive account

	## define spreadsheet title
	SPREADSHEET_TITLE = 'custom_simulations'

	## create spreadsheet if it doesn't exist
	if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
		spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE, verbose=verbose)
		folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
		GDC.move_file_to_folder(spreadsheet_id, folder_id, verbose=verbose)

	## create gspread client object
	gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)

	## run simulation
	slacgs.run()

	## upload png images to drive
	slacgs.report.upload_loss_plot_to_drive(GDC, verbose=verbose)
	slacgs.model.upload_data_points_plot_to_google_drive(GDC, verbose=verbose)

	## write results to spreadsheet
	slacgs.report.write_to_spreadsheet(gsc, dims_to_compare=dims_to_compare, verbose=verbose)

	return True


def run_experiment(start_scenario=1, verbose=True):
	""" run all simulations in all experiment scenarios

	- Reports with results will be stored in a Google Spreadsheet for each:  Experiment Scenario, Custom Experiment Scenario	and another one for the Custom Simulations.
	- The Spreadsheets are stored in a Google Drive folder named 'slacgs.demo.<user_email>'	owned by slacgs' google service	account and shared with the user's Google Drive account.
	- Also, images with data visualization will be exported to a local folder inside user's local folder (<user>/slacgs/images/ or /content/slacgs/images (for G-colab) )

	Reports Exported:
		- Loss Report: Contains mainly results focused on Loss Functions evaluations for each dimensionality of the model.
		- Compare Resport: Contains mainly results focused on comparing the performance of the Model using 2 features and 3 features.
		- Home Report (Scenario): Contains results from all simulations in a Scenario and links to the other reports. (available only for comparison between 2D and 3D)

	Images Exported (<user>/slacgs/images/ or /content/slacgs/images [for G-colab] ):
		- Scenario Data plots .gif: Contains a gif with all plots with the data points (n = 1024, dims=[2,3] ) generated for all Models in an Experiment Scenario.
		- Simulation Data plot .png: Contains a plot with the data points (n = 1024, dims=[2,3] ) generated for a Model in a Simulation.
		- Simulation Loss plot .png: Contains a plot with the loss values (Theoretical, Empirical with Train Data, Empirical with Test data) generated for a Model in a Simulation.

	Loss Functions:
		- Theoretical Loss: estimated using probability theory
		- Empirical Loss with Train Data: estimated using empirical approach with train data
		- Empirical Loss with Test Data: estimated using empirical approach with test data

	Dimensions simulated:
		- 1D: 1 feature
		- 2D: 2 features
		- 3D: 3 features

	Dimensions compared:
		- 2D vs 3D: 2 features vs 3 features

	Cardinalities simulated:
		- N = [2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]  (+ [2048, 4096, 8192] if L_3 > L_2)

	Parameters:
		start_scenario: scenario to start the experiment
		verbose: if True, prints messages to console output

	Returns:
	 		0 if all parameters have been simulated

	Raises:
		ValueError: if start_scenario is not between 1 and 4
		TypeError: if start_scenario is not an int


	Examples:
		>>> from slacgs.demo import *

		>>> # opt-1. set report service configuration with your own google cloud service account key file
		>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
		>>> set_report_service_conf(PATH_TO_KEY_FILE)

		>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
		>>> set_report_service_conf()

		>>> run_experiment()

	"""

	if not isinstance(start_scenario, int):
		raise TypeError("start_scenario must be an int")

	if start_scenario < 1 or start_scenario > 4:
		raise ValueError("start_scenario must be between 1 and 4")

	for index in range(len(SCENARIOS)):
		save_scenario_figures_as_gif(SCENARIOS[index], index + 1, verbose=verbose)

	while run_experiment_simulation(start_scenario):
		continue

	if verbose:
		print("All parameters have been simulated. Please check your google drive section: 'Shared with me' for results.")
	return 0


def doctest_next_parameter():
	""" return the next parameter to be simulated on doctests, and also the adequate spreadsheet title

	:returns: PARAM, SPREADSHEET_TITLE
	:rtype: tuple

	:Example:
	>>> from slacgs.demo import *

	>>> # opt-1. set report service configuration with your own google cloud service account key file
	>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
	>>> set_report_service_conf(PATH_TO_KEY_FILE)

	>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
	>>> set_report_service_conf()

	>>> params, spreadsheet_title = doctest_next_parameter()

	"""
	start_google_drive_service()

	REPORT_FOLDER_NAME = 'slacgs.doctest'
	SPREADSHEET_TITLE = 'scenario1.doctest'
	## create spreadsheet for the first simulation if it doesn't exist
	if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
		with contextlib.redirect_stdout(io.StringIO()):
			## do operations without printing to stdout
			spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE)
			folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
			GDC.move_file_to_folder(spreadsheet_id, folder_id)
		PARAM = SCENARIOS[0][0]
	else:  # if spreadsheet already exists, then find the first parameter that is not in the spreadsheet report home
		for i in range(len(SCENARIOS)):
			SPREADSHEET_TITLE = 'scenario' + str(i + 1) + '.doctest'

			## create spreadsheet if it doesn't exist
			if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
				spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE)
				folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
				GDC.move_file_to_folder(spreadsheet_id, folder_id)

			gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)

			## retrieve the first parameter that is not in the spreadsheet report home
			PARAM = None
			for params in SCENARIOS[i]:
				if gsc.param_not_in_home(params):
					PARAM = params
					break

			## if all parameters are in the spreadsheet report home, then go to the next spreadsheet
			if PARAM:
				break

	return PARAM, SPREADSHEET_TITLE


def run_experiment_simulation_test(start_scenario=1, verbose=True):
	""" run a simulation test for one of the experiment scenarios and return True if there are still parameters to be simulated and False otherwise.

	- Reports with results will be stored in a Google Spreadsheet for each:  Experiment Scenario, Custom Experiment Scenario	and another one for the Custom Simulations.
	- The Spreadsheets are stored in a Google Drive folder named 'slacgs.demo.<user_email>'	owned by slacgs' google service	account and shared with the user's Google Drive account.
	- Also, images with data visualization will be exported to a local folder inside user's local folder (<user>/slacgs/images/ or /content/slacgs/images (for G-colab) )

	Reports Exported:
		- Loss Report: Contains mainly results focused on Loss Functions evaluations for each dimensionality of the model.
		- Compare Resport: Contains mainly results focused on comparing the performance of the Model using 2 features and 3 features.
		- Home Report (Scenario): Contains results from all simulations in a Scenario and links to the other reports. (available only for comparison between 2D and 3D)

	Images Exported (<user>/slacgs/images/ or /content/slacgs/images [for G-colab] ):
		- Scenario Data plots .gif: Contains a gif with all plots with the data points (n = 1024, dims=[2,3] ) generated for all Models in an Experiment Scenario.
		- Simulation Data plot .png: Contains a plot with the data points (n = 1024, dims=[2,3] ) generated for a Model in a Simulation.
		- Simulation Loss plot .png: Contains a plot with the loss values (Theoretical, Empirical with Train Data, Empirical with Test data) generated for a Model in a Simulation.

	Loss Functions:
		- Theoretical Loss: estimated using probability theory
		- Empirical Loss with Train Data: estimated using empirical approach with train data
		- Empirical Loss with Test Data: estimated using empirical approach with test data

	Dimensions simulated:
		- 1D: 1 feature
		- 2D: 2 features
		- 3D: 3 features

	Dimensions compared:
		- 2D vs 3D: 2 features vs 3 features

	Cardinalities simulated:
		- N = [2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]  (+ [2048, 4096, 8192] if L_3 > L_2)

	Parameters:
		start_scenario (int): scenario to start the simulation test
		verbose (bool): if True, print simulation progress to stdout (default is True)

	Returns:
		bool: True if there are still parameters to be simulated and False otherwise

	Raises:
		ValueError: if start_scenario is not between 1 and 4
		TypeError: if start_scenario is not an int


	Example:
		>>> from slacgs.demo import *

		>>> # opt-1. set report service configuration with your own google cloud service account key file
		>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
		>>> set_report_service_conf(PATH_TO_KEY_FILE)

		>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
		>>> set_report_service_conf()

		>>> run_experiment_simulation_test(verbose=False)

	"""

	if not isinstance(start_scenario, int):
		raise TypeError("start_scenario must be an int")

	if not 1 <= start_scenario <= 4:
		raise ValueError("start_scenario must be between 1 and 4")

	start_google_drive_service()

	## define folder name for storing reports
	REPORT_FOLDER_NAME = 'slacgs.demo.' + GDC.gdrive_account_email

	## create folder if it doesn't exist
	if not GDC.folder_exists(REPORT_FOLDER_NAME):
		folder_id = GDC.create_folder(REPORT_FOLDER_NAME)  # create folder
		GDC.share_folder_with_gdrive_account(folder_id, verbose=verbose)  # share folder with user's google drive account

	SPREADSHEET_TITLE = 'scenario1.test'
	## create spreadsheet for the first simulation if it doesn't exist
	if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
		spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE, verbose=verbose)
		folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
		GDC.move_file_to_folder(spreadsheet_id, folder_id, verbose=verbose)
		gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)
		PARAM = SCENARIOS[0][0]
	else:  # if spreadsheet already exists, then find the first parameter that is not in the spreadsheet report home
		for i in range(start_scenario - 1, len(SCENARIOS)):
			SPREADSHEET_TITLE = 'scenario' + str(i + 1) + '.test'

			## create spreadsheet if it doesn't exist
			if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
				spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE, verbose=verbose)
				folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
				GDC.move_file_to_folder(spreadsheet_id, folder_id, verbose=verbose)

			gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)

			## retrieve the first parameter that is not in the spreadsheet report home
			PARAM = None
			for j in range(len(SCENARIOS[i])):
				params = SCENARIOS[i][j]
				if gsc.param_not_in_home(params):
					PARAM = params
					break

			## if all parameters are in the spreadsheet report home, then go to the next spreadsheet
			if PARAM:
				break

	if not PARAM:
		print("All parameters have been simulated. Please check your google drive section: 'Shared with me' for results.")
		return False

	## create model object
	model = Model(PARAM)

	## create simulator object
	slacgs = Simulator(model, iters_per_step=1, max_steps=10, first_step=5, precision=1e-4, augmentation_until_n=1024,
	                   verbose=verbose)

	## run simulation
	slacgs.run()

	## upload png images to drive
	slacgs.report.upload_loss_plot_to_drive(GDC, verbose=verbose)
	slacgs.model.upload_data_points_plot_to_google_drive(GDC, verbose=verbose)

	## write results to spreadsheet
	slacgs.report.write_to_spreadsheet(gsc, verbose=verbose)

	return True


def add_simulation_to_experiment_scenario_spreadsheet_test(params, scenario_number, verbose=True):
	""" add simulation results to one of the experiment scenario spreadsheets

	Parameters:
		params (list[float|int] or tuple[float|int]): a list containnong Sigmas and Rhos
		scenario_number (int): scenario number
		verbose (bool): if True, print simulation progress (default: True)

	Returns:
		None

	Raises:
		TypeError:
			if scenario is not an int;
			if params is not a list[int|float] or tuple[int|float]
		ValueError:
			if params is not a valid parameter list;
			if scenario is not between 1 and 4

	See Also:
		:func:`slacgs.demo.run_experiment_simulation_test`

	Example:
		>>> from slacgs.demo import *

		>>> # opt-1. set report service configuration with your own google cloud service account key file
		>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
		>>> set_report_service_conf(PATH_TO_KEY_FILE)

		>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
		>>> set_report_service_conf()

		>>> scenario_number = 1
		>>> params = [1, 1, 2.1, 0, 0, 0]
		>>> add_simulation_to_experiment_scenario_spreadsheet_test(params, scenario_number)

		>>> scenario_number = 2
		>>> params = [1, 1, 2, -0.15, 0, 0]
		>>> add_simulation_to_experiment_scenario_spreadsheet_test(params, scenario_number)

		>>> scenario_number = 3
		>>> params = [1, 1, 2, 0, 0.15, 0.15]
		>>> add_simulation_to_experiment_scenario_spreadsheet_test(params, scenario_number)

		>>> scenario_number = 4
		>>> params = [1, 1, 2, -0.1, 0.15, 0.15]
		>>> add_simulation_to_experiment_scenario_spreadsheet_test(params, scenario_number)


	"""

	if not isinstance(scenario_number, int):
		raise TypeError("scenario must be an int")

	if not isinstance(params, (list, tuple)):
		raise TypeError("params must be a list or tuple")

	if not all(isinstance(x, (int, float)) for x in params):
		raise TypeError("params must be a list or tuple of int or float")

	if scenario_number < 1 or scenario_number > 4:
		raise ValueError("scenario must be between 1 and 4")

	if len(params) != 6:
		raise ValueError("params must be a list or tuple of 6 elements for this experiment")

	if scenario_number == 1:
		if params[0] != 1 or params[1] != 1 or params[3] != 0 or params[4] != 0 or params[5] != 0:
			raise ValueError(
				"for scenario 1, params must be a list or tuple of 6 elements where params[0] = 1, params[1] = 1, params[3] = 0, params[4] = 0 and params[5] = 0")

	elif scenario_number == 2:
		if params[0] != 1 or params[1] != 1 or params[2] != 2 or params[4] != 0 or params[5] != 0:
			raise ValueError(
				"for scenario 2, params must be a list or tuple of 6 elements where params[0] = 1, params[1] = 1, params[2] = 2, params[4] = 0 and params[5] = 0")
		if params[3] < -0.8 or params[3] > 0.8:
			raise ValueError("for scenario 2, params[3] must be between -0.8 and 0.8")

	elif scenario_number == 3:
		if params[0] != 1 or params[1] != 1 or params[2] != 2 or params[3] != 0:
			raise ValueError(
				"for scenario 3, params must be a list or tuple of 6 elements where params[0] = 1, params[1] = 1, params[2] = 2 and params[3] = 0")
		if params[4] < -0.7 or params[4] > 0.7 or params[4] != params[5]:
			raise ValueError(
				"for scenario 3, params[4] must be between -0.7 and 0.7 and params[4] must be equal to params[5]")

	elif scenario_number == 4:
		if params[0] != 1 or params[1] != 1 or params[2] != 2 or params[3] != -0.1:
			raise ValueError(
				"for scenario 4, params must be a list or tuple of 6 elements where params[0] = 1, params[1] = 1, params[2] = 2 and params[3] = -0.1")
		if params[4] < -0.6 or params[4] > 0.6 or params[4] != params[5]:
			raise ValueError(
				"for scenario 4, params[4] must be between -0.6 and 0.6 and params[4] must be equal to params[5]")

	## update scenario gif
	save_scenario_figures_as_gif([params], scenario_number, verbose=verbose)

	start_google_drive_service()

	## define folder name for storing reports
	REPORT_FOLDER_NAME = 'slacgs.demo.' + GDC.gdrive_account_email

	## create folder if it doesn't exist
	if not GDC.folder_exists(REPORT_FOLDER_NAME):
		folder_id = GDC.create_folder(REPORT_FOLDER_NAME, verbose=verbose)  # create folder
		GDC.share_folder_with_gdrive_account(folder_id, verbose=verbose)  # share folder with user's google drive account

	## define spreadsheet title
	SPREADSHEET_TITLE = 'scenario' + str(scenario_number) + '.test'

	## create spreadsheet if it doesn't exist
	if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
		spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE, verbose=verbose)
		folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
		GDC.move_file_to_folder(spreadsheet_id, folder_id, verbose=verbose)

	## create gspread client object
	gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)

	## create model object
	model = Model(params)

	## create simulator object
	slacgs = Simulator(model, iters_per_step=1, max_steps=10, first_step=5, precision=1e-4, augmentation_until_n=1024,
	                   verbose=verbose)

	## run simulation
	slacgs.run()

	## upload png images to drive
	slacgs.report.upload_loss_plot_to_drive(GDC, verbose=verbose)
	slacgs.model.upload_data_points_plot_to_google_drive(GDC, verbose=verbose)

	## write results to spreadsheet
	slacgs.report.write_to_spreadsheet(gsc, verbose=verbose)


def run_custom_scenario_test(scenario_list, scenario_number, dims_to_simulate=None, dims_to_compare=None, verbose=True):
	""" run a custom test scenario and write the results to a Google Spreadsheet shared with the user. A Scenario is a list with params to simulate.

	- Reports with results will be stored in a Google Spreadsheet for each:  Experiment Scenario, Custom Experiment Scenario	and another one for the Custom Simulations.
	- The Spreadsheets are stored in a Google Drive folder named 'slacgs.demo.<user_email>'	owned by slacgs' google service	account and shared with the user's Google Drive account.
	- Also, images with data visualization will be exported to a local folder inside user's local folder (<user>/slacgs/images/ or /content/slacgs/images (for G-colab) )

	Reports Exported:
		- Loss Report: Contains mainly results focused on Loss Functions evaluations for each dimensionality of the model.
		- Compare Resport: Contains mainly results focused on comparing the performance of the Model using 2 features and 3 features.
		- Home Report (Scenario): Contains results from all simulations in a Scenario and links to the other reports. (available only for comparison between 2D and 3D)

	Images Exported (<user>/slacgs/images/ or /content/slacgs/images [for G-colab] ):
		- Scenario Data plots .gif: Contains a gif with all plots with the data points (n = 1024, dims=[2,3] ) generated for all Models in an Experiment Scenario.
		- Simulation Data plot .png: Contains a plot with the data points (n = 1024, dims=[2,3] ) generated for a Model in a Simulation.
		- Simulation Loss plot .png: Contains a plot with the loss values (Theoretical, Empirical with Train Data, Empirical with Test data) generated for a Model in a Simulation.

	Loss Functions:
		- Theoretical Loss: estimated using probability theory
		- Empirical Loss with Train Data: estimated using empirical approach with train data
		- Empirical Loss with Test Data: estimated using empirical approach with test data

	Cardinalities simulated:
		- N = [2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]  (+ [2048, 4096, 8192] if L_3 > L_2)

	Parameters:
		scenario_list (list[list[float|int]] or tuple[list[float|int]]): scenario list
		scenario_number (int): scenario number
		dims_to_simulate (list[int] or tuple[int]): dimensionalities to simulate (if None, all possible dimensionalities will be simulated) (default: None)
		dims_to_compare (list[int] or tuple[int]): dimensionalities to compare (if None, the last two dimensionalities of dims_to_simulate will be compared) (default: None)
		verbose (bool): whether to print messages to stdout or not

	Returns:
		bool: True if simulation was successful

	Raises:
		TypeError:
			if scenario_list is not a list[list[float|int]] or tuple[list[float|int]];
			if scenario_number is not an int;
			if dims_to_simulate is not a list[int] or tuple[int];
			if dims_to_compare is not a list[int] or tuple[int];

		ValueError:
			if scenario_number is not a valid scenario number;
			if dims_to_compare is not a subset of dims_to_simulate;
			if dims_to_compare length is not 2;


	Example:
		>>> from slacgs.demo import *

		>>> # opt-1. set report service configuration with your own google cloud service account key file
		>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
		>>> set_report_service_conf(PATH_TO_KEY_FILE)

		>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
		>>> set_report_service_conf()

		>>> scenario_list = [[1,1,3,round(0.1*rho,1),0,0] for rho in range(-1,2)]
		>>> scenario_number = 5
		>>> run_custom_scenario_test(scenario_list, scenario_number)
	"""

	if not isinstance(scenario_list, (list, tuple)):
		raise TypeError("scenario must be a list or tuple")

	if not all(isinstance(x, (list, tuple)) for x in scenario_list):
		raise TypeError("scenario must be a list or tuple of list or tuple")

	if not all(isinstance(x, (int, float)) for y in scenario_list for x in y):
		raise TypeError("scenario must be a list or tuple of list or tuple of int or float")

	if not isinstance(scenario_number, int):
		raise TypeError("scenario_number must be an int")

	if dims_to_simulate and not isinstance(dims_to_simulate, (list, tuple)):
		raise TypeError("dims_to_simulate must be a list or tuple")

	if dims_to_simulate and not all(isinstance(x, int) for x in dims_to_simulate):
		raise TypeError("dims_to_simulate must be a list or tuple of int")

	if dims_to_compare and not isinstance(dims_to_compare, (list, tuple)):
		raise TypeError("dims_to_compare must be a list or tuple")

	if dims_to_compare and not all(isinstance(x, int) for x in dims_to_compare):
		raise TypeError("dims_to_compare must be a list or tuple of int")

	if scenario_number < 5:
		raise ValueError("Custom scenario_number must be >= 5")

	start_google_drive_service()

	## create Model objects to test each parameter set before continuing
	models = []
	for params in scenario_list:
		models.append(Model(params))

	## create Simulator objects to test each parameter set before continuing
	simulators = []
	for model in models:
		simulators.append(
			Simulator(model, dims=dims_to_simulate, dims_to_compare=dims_to_compare, iters_per_step=1, max_steps=10,
			          first_step=5, precision=1e-4,
			          augmentation_until_n=1024, verbose=verbose))

	save_scenario_figures_as_gif(scenario_list, scenario_number, verbose=verbose)

	if dims_to_compare and not all(dim in dims_to_simulate for dim in dims_to_compare):
		raise ValueError("dims_to_compare must be a subset of dims_to_simulate")

	if dims_to_compare and len(dims_to_compare) != 2:
		raise ValueError("dims_to_compare length must be 2")

	## define folder name for storing reports
	REPORT_FOLDER_NAME = 'slacgs.demo.' + GDC.gdrive_account_email

	## create folder if it doesn't exist
	if not GDC.folder_exists(REPORT_FOLDER_NAME):
		folder_id = GDC.create_folder(REPORT_FOLDER_NAME, verbose=verbose)  # create folder
		GDC.share_folder_with_gdrive_account(folder_id, verbose=verbose)  # share folder with user's google drive account

	## define spreadsheet title
	SPREADSHEET_TITLE = 'custom_scenario' + str(scenario_number) + '.test'

	## create spreadsheet if it doesn't exist
	if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
		spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE, verbose=verbose)
		folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
		GDC.move_file_to_folder(spreadsheet_id, folder_id, verbose=verbose)

	## create gspread client object
	gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)

	for slacgs in simulators:
		if dims_to_compare == (2, 3) or dims_to_compare == [2, 3]:
			if gsc.param_not_in_home(slacgs.model.params):
				## run simulation
				slacgs.run()

				## upload png images to drive
				slacgs.report.upload_loss_plot_to_drive(GDC, verbose=verbose)
				slacgs.model.upload_data_points_plot_to_google_drive(GDC, verbose=verbose)

				## write results to spreadsheet
				slacgs.report.write_to_spreadsheet(gsc, verbose=verbose)
			else:
				continue

		else:
			## run simulation
			slacgs.run()

			## upload png images to drive
			slacgs.report.upload_loss_plot_to_drive(GDC, verbose=verbose)
			slacgs.model.upload_data_points_plot_to_google_drive(GDC, verbose=verbose)

			## write results to spreadsheet
			slacgs.report.write_to_spreadsheet(gsc, verbose=verbose)

	return True


def add_simulation_to_custom_scenario_spreadsheet_test(params, scenario_number, dims_to_simulate=None,
                                                       dims_to_compare=None, verbose=True):
	""" add a simulation to a custom test scenario

	Parameters:
		params (list[float|int] or tuple[float|int]): a list containnong Sigmas and Rhos
		scenario_number (int): scenario number
		dims_to_simulate (list[int] or tuple[int]): dimensionalities to simulate (if None, all dimensionalities will be simulated) (default: None)
		dims_to_compare (list[int] or tuple[int]): dimensionalities to compare (if None, the last two dimensionalities of dims_to_simulate will be compared) (default: None)
		verbose (bool): whether to print progress to console output (default: True)

	Returns:
		bool: True if successful, False otherwise

	Raises:
		TypeError:
			if params is not a list[float|int] or tuple[float|int];
			if scenario_number is not an int;
			if dims_to_simulate is not a list[int] or tuple[int];
			if dims_to_compare is not a list[int] or tuple[int]

		ValueError:
			if scenario_number is < 5;
			if dims_to_compare is not a subset of dims_to_simulate;
			if dims_to_compare length is not 2


	Example:

		>>> from slacgs.demo import *

		>>> # opt-1. set report service configuration with your own google cloud service account key file
		>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
		>>> set_report_service_conf(PATH_TO_KEY_FILE)

		>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
		>>> set_report_service_conf()

		>>> params = (1, 1, 3, -0.2, 0, 0)
		>>> scenario_number = 5
		>>> add_simulation_to_custom_scenario_spreadsheet_test(params, scenario_number)

	"""

	if not isinstance(params, (list, tuple)):
		raise TypeError("params must be a list or tuple")

	if not isinstance(scenario_number, int):
		raise TypeError("scenario_number must be an int")

	if dims_to_simulate and not isinstance(dims_to_simulate, (list, tuple)):
		raise TypeError("dims_to_simulate must be a list or tuple")

	if dims_to_compare and not isinstance(dims_to_compare, (list, tuple)):
		raise TypeError("dims_to_compare must be a list or tuple")

	if scenario_number < 5:
		raise ValueError("scenario_number must be >= 5 for a Custom Scenario")

	if dims_to_compare and dims_to_simulate and not set(dims_to_compare).issubset(set(dims_to_simulate)):
		raise ValueError("dims_to_compare must be a subset of dims_to_simulate")

	if dims_to_compare and len(dims_to_compare) != 2:
		raise ValueError("dims_to_compare length must be 2")

	## update scenario gif
	save_scenario_figures_as_gif([params], scenario_number, verbose=verbose)

	start_google_drive_service()

	## create Model object to test parameter set before continuing
	model = Model(params)

	## create Simulator object to test parameters before continuing
	slacgs = Simulator(model, dims=dims_to_simulate, dims_to_compare=dims_to_compare, iters_per_step=1, max_steps=10,
	                   first_step=5, precision=1e-4,
	                   augmentation_until_n=1024, verbose=verbose)

	## define folder name for storing reports
	REPORT_FOLDER_NAME = 'slacgs.demo.' + GDC.gdrive_account_email

	## create folder if it doesn't exist
	if not GDC.folder_exists(REPORT_FOLDER_NAME):
		folder_id = GDC.create_folder(REPORT_FOLDER_NAME, verbose=verbose)  # create folder
		GDC.share_folder_with_gdrive_account(folder_id, verbose=verbose)  # share folder with user's google drive account

	## define spreadsheet title
	SPREADSHEET_TITLE = 'custom_scenario' + str(scenario_number) + '.test'

	## create spreadsheet if it doesn't exist
	if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
		spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE)
		folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
		GDC.move_file_to_folder(spreadsheet_id, folder_id, verbose=verbose)

	## create gspread client object
	gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)

	## run simulation
	slacgs.run()

	## upload png images to drive
	slacgs.report.upload_loss_plot_to_drive(GDC, verbose=verbose)
	slacgs.model.upload_data_points_plot_to_google_drive(GDC, verbose=verbose)

	## write results to spreadsheet
	slacgs.report.write_to_spreadsheet(gsc, verbose=verbose)

	return True


def run_custom_simulation_test(params, dims_to_simulate=None, dims_to_compare=None, verbose=True):
	""" run a custom simulation for any dimensionality and cardinality

	- Reports with results will be stored in a Google Spreadsheet for each:  Experiment Scenario, Custom Experiment Scenario	and another one for the Custom Simulations.
	- The Spreadsheets are stored in a Google Drive folder named 'slacgs.demo.<user_email>'	owned by slacgs' google service	account and shared with the user's Google Drive account.
	- Also, images with data visualization will be exported to a local folder inside user's local folder (<user>/slacgs/images/ or /content/slacgs/images (for G-colab) )

	Reports Exported:
		- Loss Report: Contains mainly results focused on Loss Functions evaluations for each dimensionality of the model.
		- Compare Resport: Contains mainly results focused on comparing the performance of the Model using 2 features and 3 features.
		- Home Report (Scenario): Contains results from all simulations in a Scenario and links to the other reports. (available only for comparison between 2D and 3D)

	Images Exported (<user>/slacgs/images/ or /content/slacgs/images [for G-colab] ):
		- Scenario Data plots .gif: Contains a gif with all plots with the data points (n = 1024, dims=[2,3] ) generated for all Models in an Experiment Scenario.
		- Simulation Data plot .png: Contains a plot with the data points (n = 1024, dims=[2,3] ) generated for a Model in a Simulation.
		- Simulation Loss plot .png: Contains a plot with the loss values (Theoretical, Empirical with Train Data, Empirical with Test data) generated for a Model in a Simulation.

	Loss Functions:
		- Theoretical Loss: estimated using probability theory
		- Empirical Loss with Train Data: estimated using empirical approach with train data
		- Empirical Loss with Test Data: estimated using empirical approach with test data

	Cardinalities simulated:
		- N = [2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]  (+ [2048, 4096, 8192] if L_3 > L_2)

	Parameters:
		params (list[float|int] or tuple[float|int]): a list containnong Sigmas and Rhos
		dims_to_simulate (list[int] or tuple[int]): dimensionalities to simulate (if None, all dimensionalities will be simulated) (default: None)
		dims_to_compare (list[int] or tuple[int]): dimensionalities to compare (if None, the last two dimensionalities of dims_to_simulate will be compared) (default: None)
		verbose (bool): whether to print progress to console output (default: True)

	Returns:
		bool: True if successful, False otherwise

	Raises:
		TypeError:
			if params is not a list[float|int] or tuple[float|int];
			if dims_to_compare is not a list[int] or tuple[int]

		ValueError:
			if dims_to_compare length is not 2;
			if dims_to_compare is not a subset of dims_to_simulate


	:Example:
		>>> from slacgs.demo import *

		>>> # opt-1. set report service configuration with your own google cloud service account key file
		>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
		>>> set_report_service_conf(PATH_TO_KEY_FILE)

		>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
		>>> set_report_service_conf()

		>>> ## 2 features
		>>> params = [1, 2, 0.4]
		>>> run_custom_simulation_test(params, dims_to_compare)

		>>> ## 3 features
		>>> params = [1, 1, 4, -0.2, 0.1, 0.1]
		>>> run_custom_simulation_test(params, dims_to_compare)

		>>> ## 4 features
		>>> params = [1, 1, 1, 2, 0, 0, 0, 0, 0, 0]
		>>> dims_to_compare = (3, 4)
		>>> run_custom_simulation_test(params, dims_to_compare)

		>>> ## 5 features
		>>> params = [1, 1, 2, 2, 2, -0.3, -0.2, -0.1, 0, 0.1, 0.2, 0.2, 0, 0, 0]
		>>> dims_to_compare = (2, 5)
		>>> run_custom_simulation_test(params, dims_to_compare)

		>>> params = [1, 2, 3, 4, 5, 6, -0.3, -0.3, -0.2, -0.2, -0.1, -0.1, 0, 0, 0.1, 0.1, 0.2, 0.2, 0.3, 0.3, 0.4]
		>>> run_custom_simulation_test(params)


	"""

	if not isinstance(params, (list, tuple)):
		raise TypeError("params must be a list or tuple")

	if not all(isinstance(x, (int, float)) for x in params):
		raise TypeError("params must be a list or tuple of int or float")

	if dims_to_compare and not isinstance(dims_to_compare, (list, tuple)):
		raise TypeError("dims_to_compare must be a list or tuple")

	if dims_to_compare and not all(isinstance(x, int) for x in dims_to_compare):
		raise TypeError("dims_to_compare must be a list or tuple of int")

	if dims_to_compare and len(dims_to_compare) != 2:
		raise ValueError("dims_to_compare must be a list or tuple of length 2")

	if dims_to_compare and dims_to_simulate and not set(dims_to_compare).issubset(set(dims_to_simulate)):
		raise ValueError("dims_to_compare must be a subset of dims_to_simulate")

	## initialize gdrive client if it hasn't been initialized yet
	start_google_drive_service()

	## create model object
	model = Model(params)

	## create simulator object
	slacgs = Simulator(model, dims=dims_to_simulate, dims_to_compare=dims_to_compare, iters_per_step=1, max_steps=10,
	                   first_step=5, precision=1e-4,
	                   augmentation_until_n=1024, verbose=verbose)

	## define folder name for storing reports
	REPORT_FOLDER_NAME = 'slacgs.demo.' + GDC.gdrive_account_email

	## create folder if it doesn't exist
	if not GDC.folder_exists(REPORT_FOLDER_NAME):
		folder_id = GDC.create_folder(REPORT_FOLDER_NAME, verbose=verbose)  # create folder
		GDC.share_folder_with_gdrive_account(folder_id, verbose=verbose)  # share folder with user's google drive account

	## define spreadsheet title
	SPREADSHEET_TITLE = 'custom_simulations.test'

	## create spreadsheet if it doesn't exist
	if not GDC.check_spreadsheet_existence(SPREADSHEET_TITLE):
		spreadsheet_id = GDC.create_spreadsheet(SPREADSHEET_TITLE, verbose=verbose)
		folder_id = GDC.get_folder_id_by_name(REPORT_FOLDER_NAME)
		GDC.move_file_to_folder(spreadsheet_id, folder_id, verbose=verbose)

	## create gspread client object
	gsc = GspreadClient(report_service_conf['pygsheets_service'], SPREADSHEET_TITLE)

	## run simulation
	slacgs.run()

	## upload png images to drive
	slacgs.report.upload_loss_plot_to_drive(GDC, verbose=verbose)
	slacgs.model.upload_data_points_plot_to_google_drive(GDC, verbose=verbose)

	## write results to spreadsheet
	slacgs.report.write_to_spreadsheet(gsc, dims_to_compare=dims_to_compare, verbose=verbose)

	return True


def run_experiment_test(start_scenario=1, verbose=True):
	""" run all simulations in all experiment scenarios

	- Reports with results will be stored in a Google Spreadsheet for each:  Experiment Scenario, Custom Experiment Scenario	and another one for the Custom Simulations.
	- The Spreadsheets are stored in a Google Drive folder named 'slacgs.demo.<user_email>'	owned by slacgs' google service	account and shared with the user's Google Drive account.
	- Also, images with data visualization will be exported to a local folder inside user's local folder (<user>/slacgs/images/ or /content/slacgs/images (for G-colab) )

	Reports Exported:
		- Loss Report: Contains mainly results focused on Loss Functions evaluations for each dimensionality of the model.
		- Compare Resport: Contains mainly results focused on comparing the performance of the Model using 2 features and 3 features.
		- Home Report (Scenario): Contains results from all simulations in a Scenario and links to the other reports. (available only for comparison between 2D and 3D)

	Images Exported (<user>/slacgs/images/ or /content/slacgs/images [for G-colab] ):
		- Scenario Data plots .gif: Contains a gif with all plots with the data points (n = 1024, dims=[2,3] ) generated for all Models in an Experiment Scenario.
		- Simulation Data plot .png: Contains a plot with the data points (n = 1024, dims=[2,3] ) generated for a Model in a Simulation.
		- Simulation Loss plot .png: Contains a plot with the loss values (Theoretical, Empirical with Train Data, Empirical with Test data) generated for a Model in a Simulation.

	Loss Functions:
		- Theoretical Loss: estimated using probability theory
		- Empirical Loss with Train Data: estimated using empirical approach with train data
		- Empirical Loss with Test Data: estimated using empirical approach with test data

	Dimensions simulated:
		- 1D: 1 feature
		- 2D: 2 features
		- 3D: 3 features

	Dimensions compared:
		- 2D vs 3D: 2 features vs 3 features

	Cardinalities simulated:
		- N = [2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]  (+ [2048, 4096, 8192] if L_3 > L_2)

	Parameters:
		start_scenario: scenario to start the experiment test
		verbose: if True, prints messages to console output

	Returns:
	 		0 if all parameters have been simulated

	Raises:
		ValueError: if start_scenario is not between 1 and 4
		TypeError: if start_scenario is not an int


	Examples:
		>>> from slacgs.demo import *

		>>> # opt-1. set report service configuration with your own google cloud service account key file
		>>> PATH_TO_KEY_FILE = '/path/to/your/google/cloud/service/account/key/file.json'
		>>> set_report_service_conf(PATH_TO_KEY_FILE)

		>>> # opt-2 set report service configuration to use slacgs' server if you have the access password
		>>> set_report_service_conf()

		>>> run_experiment_test()

	"""

	if not isinstance(start_scenario, int):
		raise TypeError("start_scenario must be an int")

	if start_scenario < 1 or start_scenario > 4:
		raise ValueError("start_scenario must be between 1 and 4")

	for index in range(len(SCENARIOS)):
		save_scenario_figures_as_gif(SCENARIOS[index], index + 1, verbose=verbose)

	while run_experiment_simulation_test(start_scenario):
		continue

	if verbose:
		print("All parameters have been simulated. Please check your google drive section: 'Shared with me' for results.")
	return 0


def save_scenario_figures_as_gif(scenario, scenario_number, export_path=None, duration=200, loop=0, verbose=True):
	"""
	Save a list of matplotlib Figure objects as an animated GIF.

	- images with data visualization will be exported to a local folder inside user's local folder (<user>/slacgs/images/ or /content/slacgs/images (for G-colab) )

	Images Exported (<user>/slacgs/images/ or /content/slacgs/images [for G-colab] ):
		- Scenario Data plots .gif: Contains a gif with all plots with the data points (n = 1024, dims=[2,3] ) generated for all Models in an Experiment Scenario.


	Parameters:
		scenario (list[tuple[float|int]]): A list of parameter sets to simulate.
		scenario_number (int): The scenario number.
		export_path (str): The file path where the animated GIF will be saved.
		duration (int, optional): The duration (in milliseconds) between frames. Default is 200ms.
		loop (int, optional): The number of loops for the animation. Use 0 for infinite looping (default).

	Returns:
			None
	"""
	# Ensure the export path has the ".gif" extension
	if export_path is None:
		export_path = get_grandparent_folder_path()
		export_path += '\\images\\' if os.name == 'nt' else '/images/'
		export_path += 'scenario' + str(scenario_number) + '.gif'
	elif not export_path.endswith(".gif"):
		export_path += ".gif"

	# Get the list of figure objects
	param_figures_list = [(model.params, model.data_points_plot) for model in [Model(params) for params in scenario]]

	# Create a temporary directory to store the individual frame images
	scenario_figs_dir = get_grandparent_folder_path()
	scenario_figs_dir += '\\images\\' if os.name == 'nt' else '/images/'
	scenario_figs_dir += 'scenario' + str(scenario_number) + '_figures'
	try:
		os.makedirs(scenario_figs_dir)
	except OSError:
		pass

	try:
		# Save each figure as an individual frame image
		for i in range(len(param_figures_list)):
			params = param_figures_list[i][0]
			fig = param_figures_list[i][1]
			frame_path = os.path.join(scenario_figs_dir, str(params) + '.png')
			if not os.path.exists(frame_path):
				fig.savefig(frame_path, format="png", dpi=300)
			plt.close(fig)

		# Create the animated GIF from the frame images
		frames = [Image.open(os.path.join(scenario_figs_dir, f)) for f in os.listdir(scenario_figs_dir) if
		          f.endswith(".png")]
		frames[0].save(export_path, format="GIF", append_images=frames[1:], save_all=True, duration=duration, loop=loop)

		if verbose:
			print(f"Animated GIF saved as: {export_path}")
	except Exception as e:
		print(f"Failed to save the animated GIF: {e}")


def print_experiment_scenarios():
	""" Print the experiment scenarios parameters in a table. """

	## fill lists with empty strings to make them the same length
	def fill_lists(lst, fill_value=''):
		max_len = max(len(sublist) for sublist in lst)

		# Fill the shorter lists
		for sublist in lst:
			sublist.extend([fill_value] * (max_len - len(sublist)))

		return lst

	## convert to numpy array, transpose, convert back to list
	data = np.array(fill_lists(SCENARIOS), dtype=object).T.tolist()

	## add index column
	indexed_data = [[i] + sublist for i, sublist in enumerate(data)]

	## make table and print
	table = tabulate(indexed_data, tablefmt='grid', headers=['Scenario 1', 'Scenario 2', 'Scenario 3', 'Scenario 4'])
	print(table)



import os
from math import sqrt

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.patches import Ellipse


from .enumtypes import DictionaryType
from .utils import report_service_conf


class Model:
	"""Represents a Model for this Simulator for Loss Analysis of a Classifier."""


	def __init__(self, params, max_n=int(2 ** 13), N=tuple([int(2 ** i) for i in range(1, 11)]), dictionary=('LINEAR',)):
		"""This Model for SLACGS contains:
			- :math:`d`: dimensionality of the Model
			- :math:`\mathbf{\sigma} = \\bigcup_{i=1}^{d} \sigma_i` : list of standard deviations for each feature
			- :math:`\mathbf{\\rho} = \\bigcup_{i=1}^{d} \\bigcup_{j=i+1}^{d}  \\rho_{ij}`: list of correlations between each pair of features
			- :math:`\mathbf{N} = \\bigcup_{i=1}^{k} 2^i`, where :math:`k` is the length of :math:`\mathbf{N}` : list of cardinalities of the model
			- :math:`H`: dictionary of classifiers

		Parameters:
			params (list of numbers): list containing the standard deviation vector :math:`\mathbf{\sigma}` and the correlation vector :math:`\mathbf{\\rho}`, formally :math:`\mathbf{\sigma} \cup \mathbf{\\rho}`
			max_n (int): upper bound cardinality for the set :math:`\mathbf{N}`

		Raises:
			ValueError:
				if length of params is less than 3;
				if the length of params is not equal to the sum of the natural numbers from 1 to dim (dim = 2,3,4,...);
				if max_n is not a power of 2;
				if N is not a list of powers of 2;
				if dictionary is not a list of strings and is equal to ['linear'];
				if self.cov is not a positive definite matrix;
				if self.cov is not a symmetric matrix;
				if Sigma's are not positive numbers;
				if Rho's are not numbers between -1 and 1
				if dictionary is not a valid list of strings (see enumtypes.py for valid strings);
				if abs(rho_13) is not smaller than sqrt((1 + rho_12) / 2)

			TypeError:
				if params is not a list of numbers (floats or ints) or tuple of numbers (floats or ints);
				if max_n is not an int;
				if N is not a list of ints;
				if dictionary is not a list of strings (see enumtypes.py for valid strings);

		Example:
			>>> model = Model([1, 1, 2, 0, 0, 0])
			>>> model.save_data_points_plot_as_png()

			>>> model = Model([1, 1, 2, 0.5, 0, 0])
			>>> model.save_data_points_plot_as_png()

			>>> model = Model([1, 1, 2, 0, 0.3, 0.3])
			>>> model.save_data_points_plot_as_png()

			>>> model = Model([1, 1, 2, -0.2, -0.5, -0.5])
			>>> model.save_data_points_plot_as_png()

			>>> model = Model([1, 1, 1, -0.1, 0.5, 0.5], max_n=2**15, N=[2**i for i in range(1,14)])
			>>> model.save_data_points_plot_as_png()

			>>> model = Model([1, 2, 4, 0, 0.5, 0.5], max_n=2**10, N=[2**i for i in range(1,11)])
			>>> model.save_data_points_plot_as_png()

			>>> model = Model([1, 1, 1, 2, 0.1, 0, 0, 0, 0, 0])
			>>> model.save_data_points_plot_as_png()

			>>> model = Model([1, 2, -0.1])
			>>> model.save_data_points_plot_as_png()

		"""
		if not isinstance(params, list) and not isinstance(params, tuple):
			raise TypeError('params must be a list or tuple of numbers (floats or ints)')

		if not all(isinstance(param, int) or isinstance(param, float) for param in params):
			raise TypeError('params must be a list or tuple of numbers (floats or ints)')

		if not isinstance(max_n, int):
			raise TypeError('max_n must be an int')

		if not isinstance(N, list) and not isinstance(N, tuple):
			raise TypeError('N must be a list or tuple of ints')

		if not isinstance(dictionary, list) and not isinstance(dictionary, tuple):
			raise TypeError('dictionary must be a list or tuple of strings')

		if len(params) < 3:
			raise ValueError('Check parameters list lenght, this experiment requires at least 3 parameters (case dim = 2)')

		dim = 2
		param_len = 3
		while param_len < len(params):
			dim += 1
			param_len += dim

		if param_len > len(params):
			raise ValueError('Check parameters list lenght')

		for d in range(dim):
			if params[d] <= 0:
				raise ValueError('Every Sigma must be a positive number')

		for d in range(dim, len(params)):
			if params[d] < -1 or params[d] > 1:
				raise ValueError('Every Rho must be a number between -1 and 1')

		if max_n & (max_n - 1) != 0:
			raise ValueError('max_n must be a power of 2')

		for n in N:
			if n & (n - 1) != 0:
				raise ValueError('N must be a list of powers of 2 to make this experiment')

		params = list(params)

		self.dim = dim
		self.sigma = params[0:dim]
		self.rho = params[dim:len(params)]

		if self.dim > 2:
			if not abs(self.rho[1]) < sqrt((1 + self.rho[0]) / 2):
				raise ValueError('abs(rho_13) must be smaller than sqrt((1 + rho_12) / 2)')

		self.mean_pos = [1 for d in range(dim)]
		self.mean_neg = [-1 for d in range(dim)]
		self.dictionary = list(dictionary)

		summ = 0
		aux1 = [summ]
		for i in range(1, len(self.sigma) - 1):
			summ += len(self.sigma) - i
			aux1.append(summ)

		summ = len(self.sigma) - 1
		aux2 = [summ]
		for i in range(1, len(self.sigma) - 1):
			summ += len(self.sigma) - (i + 1)
			aux2.append(summ)

		self.rho_matrix = [[None] * (i + 1) + self.rho[aux1[i]:aux2[i]] for i in range(len(self.sigma) - 1)]
		self.params = params
		self.N = list(N)
		self.max_n = max_n

		self.cov = [[self.sigma[p] ** 2 if p == q else self.sigma[p] * self.sigma[q] * self.rho_matrix[p][q] if q > p else
		self.sigma[p] * self.sigma[q] * self.rho_matrix[q][p] for q in range(len(self.sigma))] for p in
		            range(len(self.sigma))]

		if not np.all(np.linalg.eigvals(self.cov) > 0):
			raise ValueError('cov must be a positive definite matrix to make this experiment, check your parameters')

		if not np.allclose(self.cov, np.array(self.cov).T):
			raise ValueError('cov must be a symmetric matrix to make this experiment, check your parameters')

		if not all(dictionary in DictionaryType.__members__ for dictionary in dictionary):
			raise ValueError('invalid dictionary, implemented dictionaries are: ' + ', '.join(DictionaryType.__members__))

		self.data_points_plot = self.plot_surrounding_ellipsis_and_ellipsoids() if dim == 3 else None

	def save_data_points_plot_as_png(self, export_path=None, verbose=True):
		"""
		Save a matplotlib Figure object as a PNG image.

		Parameters:
			export_path (str): The file path where the PNG image will be saved.
			verbose (bool): If True, print the export path.

		Returns:
				None

		Warning:
			this method is available only for 3D models.

		"""
		if self.data_points_plot is not None:
			if export_path is None:
				export_path = report_service_conf['images_path']
				export_path +=  'data_points' +  str(self.params) + '.png'
			elif not export_path.endswith(".png"):
				export_path = report_service_conf['images_path']
				export_path += 'data_points' + str(self.params) + '.png'

			if not os.path.exists(export_path):
				try:
					self.data_points_plot.savefig(export_path, format="png", dpi=300)
					if verbose:
						print(f"Figure saved as: {export_path}")
				except Exception as e:
					print(f"Failed to save the figure: {e}")
			else:
				if verbose:
					print(f"File already exists: {export_path}")

		else:
			if verbose:
				print("No figure to save.")

		return export_path
	def upload_data_points_plot_to_google_drive(self, gdc, verbose=True):
		"""
		Upload a matplotlib Figure object as a PNG image to Google Drive.

		Parameters:
			gdc (GoogleDriveClient): The Google Drive client object.
			verbose (bool): If True, print the export path.

		Returns:
				None

		Warning:
			this method is available only for 3D models.

		"""
		if self.data_points_plot is not None:
			export_path = report_service_conf['images_path']
			export_path += 'data_points' + str(self.params) + '.png'

			gdrive_images_folder_path = 'slacgs.demo.' + gdc.gdrive_account_email + '/images'
			if not os.path.exists(export_path):
				self.save_data_points_plot_as_png(export_path=export_path, verbose=verbose)
			if not gdc.folder_exists_by_path(gdrive_images_folder_path):
				folder_id = gdc.create_folder('images', gdc.get_folder_id_by_name('slacgs.demo.' + gdc.gdrive_account_email), verbose=verbose)
			else:
				folder_id = gdc.get_folder_id_by_path(gdrive_images_folder_path)

			gdc.upload_file_to_drive(export_path, folder_id, verbose=verbose)
		else:
			if verbose:
				print("No figure to upload.")

	def plot_surrounding_ellipsis_and_ellipsoids(self) -> Figure:
		"""
		Plots the ellipsoids for this model's covariance matrix :math:`\Sigma` and a dataset sample with :math:`n=1024` sample points for :math:`d \in [2,3]`.

		Parameters:
			self (Model): The model object.

		Returns:
			Figure: The matplotlib Figure object.

		Raises:
			ValueError:
				if cov is not a 3x3 matrix;
				if cov is not a positive definite matrix;
				if cov is not a symmetric matrix;

		Warning:
			This method is available only for 3D models.

		Example:
			>>> from slacgs.model import Model
			>>> model = Model((1, 1, 2, 0.5, 0, 0))
			>>> plot_fig = model.data_points_plot

		"""

		cov = self.cov

		if len(cov) != 3 or len(cov[0]) != 3 or len(cov[1]) != 3 or len(cov[2]) != 3:
			raise ValueError('cov must be a 3x3 matrix to make this plot')

		if not np.all(np.linalg.eigvals(cov) > 0):
			raise ValueError('cov must be a positive definite matrix to make this plot')

		if not np.allclose(cov, np.array(cov).T):
			raise ValueError('cov must be a symmetric matrix to make this plot')

		# Define mean and covariance for 3D
		mean = [1, 1, 1]
		mean1 = [-1, -1, -1]
		covariance = np.array(cov[0:3]).T[0:3].T

		# Generate 1024 samples of bivariate Gaussian points
		points = np.random.multivariate_normal(mean, covariance, 1024)
		points1 = np.random.multivariate_normal(mean1, covariance, 1024)

		# Compute the eigenvectors and eigenvalues of the covariance matrix
		eigenvalues, eigenvectors = np.linalg.eig(covariance)

		# Sort the eigenvalues in decreasing order
		sorted_indices = eigenvalues.argsort()[::-1]
		eigenvalues = eigenvalues[sorted_indices]
		eigenvectors = eigenvectors[:, sorted_indices]

		# Compute the radii of the ellipsoid
		radii = np.sqrt(5.991 * eigenvalues)

		# Generate the ellipsoid mesh
		u = np.linspace(0, 2 * np.pi, 100)
		v = np.linspace(0, np.pi, 100)
		x = radii[0] * np.outer(np.cos(u), np.sin(v))
		y = radii[1] * np.outer(np.sin(u), np.sin(v))
		z = radii[2] * np.outer(np.ones_like(u), np.cos(v))
		ellipsoid = np.array([x.flatten(), y.flatten(), z.flatten()]).T  # reshape to (10000, 3)
		ellipsoid1 = np.array([x.flatten(), y.flatten(), z.flatten()]).T  # reshape to (10000, 3)

		# Apply rotation and translation to the ellipsoid mesh
		transformed_ellipsoid = np.dot(eigenvectors, ellipsoid.T).T
		transformed_ellipsoid += mean
		transformed_ellipsoid1 = np.dot(eigenvectors, ellipsoid1.T).T
		transformed_ellipsoid1 += mean1

		# Reshape the transformed ellipsoid mesh to (100, 100, 3)
		transformed_ellipsoid = transformed_ellipsoid.reshape((100, 100, 3))
		transformed_ellipsoid1 = transformed_ellipsoid1.reshape((100, 100, 3))

		# Plot the points and the ellipsoid
		fig = plt.figure(figsize=(10, 4))
		ax2 = fig.add_subplot(121, projection='3d')
		ax2.scatter(points[:, 0], points[:, 1], points[:, 2], alpha=0.3)
		ax2.scatter(points1[:, 0], points1[:, 1], points1[:, 2], alpha=0.3)
		ax2.plot_wireframe(transformed_ellipsoid[:, :, 0], transformed_ellipsoid[:, :, 1], transformed_ellipsoid[:, :, 2],
		                   color='b', alpha=0.6)
		ax2.plot_wireframe(transformed_ellipsoid1[:, :, 0], transformed_ellipsoid1[:, :, 1],
		                   transformed_ellipsoid1[:, :, 2], color='darkorange', alpha=0.8)
		ax2.set_xlabel('$x_1$')
		ax2.set_ylabel('$x_2$')
		ax2.set_zlabel('$x_3$')
		ax2.set_xlim3d([-10, 10])
		ax2.set_ylim3d([-10, 10])
		ax2.set_zlim3d([-10, 10])
		ax2.view_init(elev=30, azim=-45)
		plt.subplots_adjust(left=0.03, bottom=0.124, right=0.912, top=0.983, wspace=0.395, hspace=0.2)

		# Define mean and covariance for 2D
		mean = [1, 1]
		mean1 = [-1, -1]
		covariance = np.array(cov[0:2]).T[0:2].T

		# Generate 1024 samples of bivariate Gaussian points
		points = np.random.multivariate_normal(mean, covariance, 1024)
		points1 = np.random.multivariate_normal(mean1, covariance, 1024)

		# Compute the eigenvalues and eigenvectors of the covariance matrix
		eigenvalues, eigenvectors = np.linalg.eig(covariance)

		# Sort the eigenvalues in decreasing order
		sorted_indices = eigenvalues.argsort()[::-1]
		eigenvalues = eigenvalues[sorted_indices]
		eigenvectors = eigenvectors[:, sorted_indices]

		# Compute the angle of rotation
		theta = np.degrees(np.arctan2(*eigenvectors[:, 0][::-1]))

		# Compute the width and height of the ellipse
		width, height = 2 * np.sqrt(5.991 * eigenvalues)

		# Plot the points and the ellipse
		ax1 = fig.add_subplot(122)
		ax1.scatter(points[:, 0], points[:, 1], s=5, alpha=0.2)
		ax1.scatter(points1[:, 0], points1[:, 1], s=5, alpha=0.2)

		ellipse = Ellipse(xy=mean, width=width, height=height, angle=theta, fill=False, color='b', alpha=0.8)
		ax1.add_patch(ellipse)
		ellipse = Ellipse(xy=mean1, width=width, height=height, angle=theta, fill=False, color='orange', alpha=1)
		ax1.add_patch(ellipse)


		ax1.set_xlabel('$x_1$')
		ax1.set_ylabel('$x_2$')
		ax1.set_xlim([-10, 10])
		ax1.set_ylim([-10, 10])

		return fig



import os
import math
import time

import numpy as np

from matplotlib import pyplot as plt
from sklearn import svm
from scipy.stats import multivariate_normal, norm
from IPython.display import clear_output

from .enumtypes import LossType
from .model import Model
from .report import Report
from .utils import is_notebook

def cls():
  """
  Clears the terminal screen.
  """
  if os.name == 'nt':  # For Windows
    _ = os.system('cls')
  else:  # For Linux and Mac
    _ = os.system('clear')


def h_error_rate(h_bias, h_weights, cov):
  """
  This function calculates the probability of error for a linear classifier :math:`h` trained by Dataset :math:`D` generated by a Gaussian distribution, given:
    - weights vector :math:`\mathbf{w}` of the classifier :math:`h`;
    - the bias :math:`b` of the classifier :math:`h`;
    - the covariance matrix :math:`\Sigma` used to generate the trainning dataset :math:`D`.

  Linear Classifier?
    - ":math:`h(\mathbf{x}) = \mathbf{w} \cdot \mathbf{x} + b`" is a linear classifier.
    - The hyperplane  :math:`h` defined by this equation represents the decision boundary that separates classes in a linearly separable space.
    - The hypothesis space :math:`H` consists of all possible linear classifiers.

  The linear classifier function :math:`h(\mathbf{x})` returns a scalar value that represents a linear combination of the input data point :math:`\mathbf{x}` and the weight vector :math:`\mathbf{w}`, plus a bias term :math:`b`. In the context of binary classification (wich is the case in this experiment), this value can be interpreted as the raw score or decision boundary for classifying an input. If you apply a threshold, such as 0, you can make a binary decision:
    - If :math:`h(\mathbf{x}) \geq 0`, then classify the input as one class (e.g., class 1).
    - If :math:`h(\mathbf{x}) < 0`, then classify the input as the other class (e.g., class 0).


  The minimum probability of error for a linear classifier :math:`h` trained by Dataset :math:`D` generated by a Gaussian distribution is given by:
    - :math:`P(Error) = \\frac{1}{2} (1 - \Phi({Distance}(+))) + \\frac{1}{2}(1 - \Phi(Distance(-)))`
    - :math:`P(Error) = \\frac{1}{2} (1 - \Phi( \\frac{|1 + b - (w_0 + ... + w_{d-1})|}{\sqrt{\Delta}})) + \\frac{1}{2}(1 - \Phi(\\frac{|1 - b - (w_0 + ... + w_{d-1})|}{\sqrt{\Delta}}))`

  Where:
    - :math:`d` is the dimensionality of the classifier :math:`h`
    - :math:`\mathbf{w}` is the weights vector of the classifier :math:`h`
    - :math:`\mathbf{w} = [w_0, ..., w_{d-1}]`
    - :math:`b` is the bias of the classifier :math:`h`
    - :math:`\lambda` is the Lower-triangular Cholesky factor of :math:`\Sigma`
    - :math:`\mathbf{\delta} = [\delta_i]` for :math:`i \in [1,...,d] = [\delta_0,...,\delta_{d}]`
    - :math:`\mathbf{\delta} = (\mathbf{w}|[-1]) \cdot \lambda^T = [w_0, ..., w_{d-1},-1]\cdot \lambda^T`
    - :math:`\Delta = \sum_{\delta_i \in \\mathbf{\delta}} \delta_i^2 = \delta_0^2 + ... + \delta_{d}^2`
    - :math:`\Phi(\cdot)` is the cumulative distribution function of a Gaussian with zero mean and variance one.
    - :math:`Distance(+)` and :math:`Distance(-)` are the distances from the hyperplane :math:`h` to the mean of the positive and negative classes, :math:`\mathbf{\mu_+}` and :math:`\mathbf{\mu_-}`, respectively.
    - :math:`\mathbf{\mu_+}` : mean vector for class :math:`(+)`
    - :math:`\mathbf{\mu_-}` : mean vector for class :math:`(-)`
    - :math:`\mathbf{\mu_+} = \\bigcup_{i=1}^{d} \mu_{+i}`
    - :math:`\mathbf{\mu_+} = \\bigcup_{i=1}^{d} \mu_{-i}`
    - :math:`\mu_{+i} = 1` and :math:`\mu_{-i} = -1`, for :math:`i \in [1,...,d]`
    - :math:`\mathbf{\mu_-} = [\mu_{-i}]` for :math:`i \in [1,...,d] = [\mu_{-1}, ..., \mu_{-d}] = [1,...,1]`
    - :math:`\mathbf{\mu_+} = [\mu_{+i}]` for :math:`i \in [1,...,d] = [\mu_{+1}, ..., \mu_{+d}] = [-1,...,-1]`


  Parameters:
    h_bias (float): bias of the classifier :math:`h(\mathbf{x})`
    h_weights (list): weights of the classifier :math:`h(\mathbf{x})`
    cov (list[list]): covariance matrix of the gaussian samples used to train the model

  Returns:
    float: probability of error for the classifier :math:`h(\mathbf{x})` trained with the gaussian samples generated with the covariance matrix :math:`\Sigma` and means :math:`\mu_+=[1,...,1]` and :math:`\mu_=[-1,...,-1]`



  """

  lamb = np.linalg.cholesky(cov)

  h_weights_and_minus_one = h_weights + [-1]
  delta = np.sum(np.power([np.dot(h_weights_and_minus_one, np.array(lamb).T[i]) for i in range(len(h_weights_and_minus_one))], 2))

  arg1 = abs(np.sum(h_weights_and_minus_one) + h_bias) / math.sqrt(delta)
  arg2 = abs(np.sum(h_weights_and_minus_one) - h_bias) / math.sqrt(delta)
  pr1 = 1-norm.cdf(arg1)
  pr2 = 1-norm.cdf(arg2)
  pr = 0.5*(pr1+pr2)

  return pr


def loss_empirical(clf, dataset_test):
  """
  this function calculates an empirical loss :math:`\hat{L}(h)`, given:
    - SVM trained model containing the best empirical classifier :math:`\hat{h}^{(D)}` in the dictionary :math:`H` for the dataset :math:`D` with :math:`d` features and :math:`n` samples.
    - a test set :math:`D_{test} = D_{(+)test} \cup D_{(-)test}`

  For each classifier :math:`h` in :math:`H`, we can calculate its empirical error rate :math:`\hat{L}(h)`, i.e., its proportion of misclassified observations in the dataset :math:`D`:
    - :math:`\hat{L}(h) = \\frac{1}{n} \sum_{i=1}^{n} 1(y_i \\neq h(x_i))`

  Let :math:`\hat{h}^{(D)}` be the best empirical classifier in the dictionary :math:`H` for the dataset :math:`D` with :math:`d` features and :math:`n` samples, i.e., the classifier in :math:`H` with the minimum empirical error rate. Under these conditions, we can define two error rates:
    - :math:`\hat{L}(\hat{h}^{(D)}) =` the empirical error rate of the classifier :math:`\hat{h}^{(D)}`, when :math:`D_{test} = D_{train}`
    - :math:`\hat{L}(\hat{h}^{(D)},D') =` the empirical error rate of the classifier :math:`\hat{h}^{(D)}` using a test dataset :math:`D'`, when :math:`D_{test} \\neq D_{train}`

  Parameters:
    clf (sklearn.svm.SVC): SVM trained model containing the best empirical classifier :math:`\hat{h}^{(D)}` in the dictionary :math:`H` for the dataset :math:`D` with :math:`d` features and :math:`n` samples.
    dataset_test (dict): a test set :math:`D_{test} = D_{(+)test} \cup D_{(-)test}`

  Returns:
    float: empirical error rate of the classifier :math:`\hat{h}^{(D)}`, :math:`\hat{L}(\hat{h}^{(D)})` when :math:`D_{test} = D_{train}` or :math:`\hat{L}(\hat{h}^{(D)},D')` when :math:`D_{test} \\neq D_{train}`

  """

  dims_to_remove = [i for i in range(len(dataset_test['data'][0])-1,len(clf.coef_[0])-1,-1)]

  X_test, y_test = np.delete(dataset_test['data'], dims_to_remove, 1) , dataset_test['target']

  return 1 - clf.score(X_test,y_test)


def linear_svm_train(dim, training_set):
  """This function trains a linear SVM model and find the best empirical linear classifier :math:`\hat{h}^{(D)}` in the hypothesis space :math:`H` for the dataset :math:`D` with :math:`d` features and :math:`n` samples, given:
    - Training set :math:`D = D_{(+)train} \cup D_{(-)train}`
    - dimensionality :math:`d`, :math:`d` < nuber of features in the dataset

  Linear Classifier?
    - ":math:`h(\mathbf{x}) = \mathbf{w} \cdot \mathbf{x} + b`" is a linear classifier.
    - The hyperplane :math:`h` defined by this equation represents the decision boundary that separates classes in a linearly separable space.
    - The hypothesis space :math:`H` consists of all possible linear classifiers.

  :param self: this simulator
  :type self: Simulator
  :param dim: dimensionality of the model, :math:`d` features
  :type dim: int
  :param training_set: dataset :math:`D` to train the model
  :type training_set: dict
  :return: trained SVM model containing the best empirical hyperplane :math:`h` in the hypothesis space :math:`H` for the given dataset :math:`D` with :math:`d` features and :math:`n` samples
  :rtype: sklearn.svm._classes.SVC

  """

  # remove k columns from sample_data for model.dim-k features
  dims_to_remove = [i for i in range(len(training_set['data'][0]) - 1, dim - 1, -1)]

  X_train, y_train = np.delete(training_set['data'], dims_to_remove, 1) , training_set['target']

  # train
  C = 1.0  # SVM regularization parameter
  clf = svm.SVC(kernel='linear', C=C, probability=True) #set svm linear separator
  clf.fit(X_train, y_train.ravel()) #train model

  return clf


def loss_theoretical(h_clf, cov):
  """
  Using probability theory, calculates the theoretical loss for given:
    - sklearn SVM trained model containing the best linear classifier :math:`h` in the hypothesis space :math:`H` for the given training dataset :math:`D` with :math:`d` features and :math:`n` samples
    - covariance matrix of the gaussian samples used to train the SVM model

  Linear Classifier?
    - ":math:`h(\mathbf{x}) = \mathbf{w} \cdot \mathbf{x} + b`" is a linear classifier.
    - The hyperplane :math:`h` defined by this equation represents the decision boundary that separates classes in a linearly separable space.
    - The hypothesis space :math:`H` consists of all possible linear classifiers.

  Let :math:`\hat{h}^{(D)}` be the best empirical classifier in the dictionary :math:`H` for the dataset :math:`D` with :math:`d` features and :math:`n` samples, i.e., the classifier in :math:`H` with the minimum empirical error rate. Under these conditions, we can define theoretical error rate as:
    - :math:`L(\hat{h}^{(D)}) =` the theoretical error rate of the classifier :math:`\hat{h}^{(D)}`

  Parameters:
    h_clf(sklearn.svm._classes.SVC): SVM trained classifier
    cov(list[list] or np.ndarray): covariance matrix :math:`\Sigma` of the gaussian samples used to train the SVM model

  Returns:
    float: theoretical loss for classifier :math:`\hat{h}^{(D)}`

  Raises:
    TypeError:
      if h_clf is not a sklearn.svm._classes.SVC
      if cov is not a list of lists or np.ndarray

    ValueError:
      if h_clf is not a linear classifier
      if h_clf is not a trained scikit-learn SVM model
      if cov is not a positive definite matrix

  """

  if not isinstance(h_clf, svm.SVC):
    raise TypeError('h_clf must be a sklearn.svm._classes.SVC')

  if not isinstance(cov, list) and not isinstance(cov, np.ndarray):
    raise TypeError('cov must be a list of lists or np.ndarray')

  if not np.all(np.linalg.eigvals(cov) > 0):
    raise ValueError('cov must be a positive definite matrix')

  if h_clf.kernel != 'linear':
    raise ValueError('h_clf must be a linear classifier')

  if not hasattr(h_clf, 'coef_'):
    raise ValueError('h_clf must be a trained scikit-learn linear SVM model')

  coef_ = h_clf.coef_[0]

  bias = (h_clf.intercept_[0]) / coef_[len(coef_) - 1]

  weights = [-coef_[i]/coef_[len(coef_)-1] for i in range(len(coef_)-1)]

  return h_error_rate(bias, weights, cov)


def loss_bayes_analytical(cov):
  """calculate the theoretical Bayes Error Rate (or Bayes Risk) analytically, given:
    - covariance matrix :math:`\Sigma` of a Gaussian distribution

  The Bayes Error (also: Bayes Risk) is defined as the minimum theoretical achievable error rate:
    - :math:`\min_{h\in H} L(h) =` the smallest theoretically achievable error rate for a classifier :math:`h` in dictionary :math:`H`.

  Warning:
    This function is only valid for 1D, 2D and 3D gaussian models.

  Parameters:
    cov(list[list] or np.ndarray): covariance matrix :math:`\Sigma`

  Returns:
    float: The smallest theoretically achievable error rate for a classifier :math:`h` in dictionary :math:`H` for the given covariance matrix :math:`\Sigma`, also called Bayes Error Rate (or Bayes Risk), represented by :math:`\min_{h\in H} L(h)`.

  """
  # Extract sigma (standard deviations)
  sigma = np.sqrt(np.diag(cov))

  # Create the correlation matrix (rho)
  rho = cov / (sigma[:, None] * sigma[None, :])

  # Calculate the Bayes Error Rate (or Bayes Risk)
  if len(cov) == 3:
    if rho[0][1] == 0 and rho[0][2] == 0 and rho[1][2] == 0:
      w_0 = -(sigma[2]**2)/sigma[0]**2
      w_1 = -(sigma[2]**2)/sigma[1]**2

    elif sigma[0] == sigma[1] and rho[0][2] == 0 and rho[1][2] == 0 and rho[0][1] != 0:
      w_0 = w_1 = -(sigma[2]**2)/((sigma[0]**2)*(1+rho[0][1]))

    elif rho[0][1]==0 and rho[0][2] == rho[1][2] and sigma[0] == sigma[1]:
      try:
        w_0 = w_1 = sigma[2]*(sigma[2]-rho[1][2]*sigma[0])/(sigma[0]*(2*rho[0][2]*sigma[2]-sigma[0]))
      except ZeroDivisionError:
        return 0

    elif sigma[0] == sigma[1] and rho[0][1] != 0 and rho[0][2] == rho[1][2] and aw_0s(rho[0][2]) <= math.sqrt((1+rho[0][1])/2):
      w_0 = w_1 = (1-rho[0][2])/(2*rho[0][2] - (1+rho[0][1]))

    else:
      return 0

    return h_error_rate(0, [w_0, w_1], cov)

  elif len(cov) == 2:
    w_0 = (cov[0][1]-cov[1][1])/(cov[0][0] - cov[0][1])
    return h_error_rate(0, [w_0], cov)

  elif len(cov) == 1:
    return h_error_rate(0, [], cov)

  else:
    return 0


def distance_from_origin_to_intersection_between_normalized_ellipsoid_and_main_diagonal(cov):
  """
  This function calculates the distance from origin :math:`[0,0,...,0]` to the intersection point between the normalized ellipsoid and the main diagonal :math:`[1,1,...,1]`, given:
    - the covariance matrix :math:`\Sigma` of the ellipsoid

  Parameters:
    cov (list): covariance matrix :math:`\Sigma` of the ellipsoid

  Returns:
    float: distance from origin :math:`[0,0,...,0]` to the intersection point between the normalized ellipsoid and the main diagonal :math:`[1,1,...,1]`.

  """

  # convert cov to numpy array
  cov = np.array(cov)

  # Extract sigma (standard deviations)
  sigma = np.sqrt(np.diag(cov))

  # Extract dimension
  dim = len(sigma)

  # Extract min sigma
  min_sigma = min(sigma)

  # Define precision
  precision = 1e-4

  # Define support vector for t
  ar = np.arange(min_sigma, 0, step=-precision).tolist()

  line = np.array([[ar[i] for j in range(dim)] for i in range(len(ar))])

  candidates = []

  for i in range(len(ar)):
    X = line[i]
    if round(np.dot(np.dot(X,np.linalg.inv(cov)), X.T),2) == 1:
      candidates.append(X)
    elif candidates:
      intersection_point = np.mean(candidates)
      distance = math.sqrt(2)*(intersection_point)

      return distance


class Simulator:
  """A simulator for Linear classifier Loss analysis in order to evaluate Trade Off Between Samples and Features in Classification Problems on multivariate Gaussian Generated Samples."""

  def __init__(self, model: Model, dims=None, dims_to_compare=None, loss_types = ('EMPIRICAL_TRAIN', 'THEORETICAL', 'EMPIRICAL_TEST'), test_samples_amt=1024, iters_per_step=5, max_steps=200, first_step=100, precision=1e-6, augmentation_until_n = 1024, verbose=True):

    """
    :param model: Linear Classifier Loss Analysis Model
    :param dims: dimensionalities to be simulated
    :type dims: list[int] or tuple[int]
    :param loss_types: types of Loss comparation graphs wich will be compiled by the Simulator
    :type loss_types: list[str] or tuple[str]
    :param test_samples_amt: number of test samples to be genarated for prediction Loss evaluation
    :param iters_per_step: number of datasets generated in one simulation step equals iters_per_step*sqrt(augmentation_until_n)/sqrt(n), if n < augmentation_until_n, else it equals iters_per_step
    :param max_steps: max number of steps per n ∈ N
    :param first_step: min number of steps per n ∈ N
    :param precision: stop criteria. The simulation finishes when 'difference between E[predict_loss] calculated on two consecutive steps' < precision
    :param augmentation_until_n: increase number of generated datasets for n < augmentation_until_n
    :param verbose: if True, prints simulation progress

    :raise TypeError: if model is not a Model object;
                      if dims is not a list or tuple of integers;
                      if loss_types is not a list or tuple of strings;
                      if test_samples_amt is not an integer;
                      if iters_per_step is not an integer;
                      if max_steps is not an integer;
                      if first_step is not an integer;
                      if precision is not a float;
                      if augmentation_until_n is not an integer;
                      if verbose is not a boolean;


    :raise ValueError: if max_steps < first_step;
                      if precision < 0 or if precision > 1;
                      if augmentation_until_n < 1024;
                      if test_samples_amt < 2;
                      if iters_per_step < 1;
                      if max_steps < 10;
                      if first_step < 1
                      if loss_types is not a valid list or tuple of strings (see enumtypes.py for valid strings) or if loss_types is empty;
                      if dims is not a valid list or tuple of integers or if dims is empty;


    :Example:
      >>> from slacgs import Model

      >>> param = [1,1,2,0,0,0]
      >>> model = Model(param, N=[2**i for i in range(1,11)], max_n=1024)
      >>> slacgs = Simulator(model, iters_per_step=1, max_steps=10, first_step=5, precision=1e-4, augmentation_until_n = 1024, verbose=False)
      >>> slacgs.run() # doctest: +ELLIPSIS


      >>> param = [1,1,2,-0.1,0,0]
      >>> model = Model(param, N=[2**i for i in range(1,11)], max_n=1024)
      >>> slacgs = Simulator(model, iters_per_step=1, max_steps=10, first_step=5, precision=1e-4, augmentation_until_n = 1024, verbose=False)
      >>> slacgs.run() # doctest: +ELLIPSIS


      >>> param = [1,1,2,0,-0.4,-0.4]
      >>> model = Model(param, N=[2**i for i in range(1,11)], max_n=1024)
      >>> slacgs = Simulator(model, iters_per_step=1, max_steps=10, first_step=5, precision=1e-4, augmentation_until_n = 1024, verbose=False)
      >>> slacgs.run() # doctest: +ELLIPSIS


      >>> param = [1,1,2,-0.1,-0.4,-0.4]
      >>> model = Model(param, N=[2**i for i in range(1,11)], max_n=1024)
      >>> slacgs = Simulator(model, iters_per_step=1, max_steps=10, first_step=5, precision=1e-4, augmentation_until_n = 1024, verbose=False)
      >>> slacgs.run() # doctest: +ELLIPSIS


      >>> param = [1,2,-0.1]
      >>> model = Model(param, N=[2**i for i in range(1,11)], max_n=1024)
      >>> slacgs = Simulator(model, dims=(1,2), iters_per_step=1, max_steps=10, first_step=5, precision=1e-4, augmentation_until_n = 1024, verbose=False)
      >>> slacgs.run() # doctest: +ELLIPSIS


      >>> param = [1,1,1,2,0,0,0,0,0,0]
      >>> model = Model(param, N=[2**i for i in range(1,11)], max_n=1024)
      >>> slacgs = Simulator(model, dims=(3,4), iters_per_step=1, max_steps=10, first_step=5, precision=1e-4, augmentation_until_n = 1024, verbose=False)
      >>> slacgs.run() # doctest: +ELLIPSIS


    """


    if not isinstance(model, Model):
        raise TypeError('model must be a Model object')

    if dims and not isinstance(dims, list) and not isinstance(dims, tuple):
        raise TypeError('dims must be a list or tuple')

    if not isinstance(loss_types, list) and not isinstance(loss_types, tuple):
        raise TypeError('loss_types must be a list or tuple of strings')

    if not isinstance(test_samples_amt, int):
        raise TypeError('test_samples_amt must be an integer')

    if not isinstance(iters_per_step, int):
        raise TypeError('iters_per_step must be an integer')

    if not isinstance(max_steps, int):
        raise TypeError('max_steps must be an integer')

    if not isinstance(first_step, int):
        raise TypeError('first_step must be an integer')

    if not isinstance(precision, float):
        raise TypeError('precision must be a float')

    if not isinstance(augmentation_until_n, int):
        raise TypeError('augmentation_until_n must be an integer')

    if not all(isinstance(loss_type, str) for loss_type in loss_types):
        raise TypeError('loss_types must be a list of strings')

    if dims and not all(isinstance(dim, int) for dim in dims):
        raise TypeError('dims must be a list of integers')

    if not isinstance(verbose, bool):
        raise TypeError('verbose must be a boolean')

    if max_steps < first_step:
        raise ValueError('first_step must be less than max_steps')

    if precision < 0 or precision > 1:
        raise ValueError('precision must be greater than 0 and less than 1')

    if augmentation_until_n < 1024:
        raise ValueError('augmentation_until_n must be greater than 1024')

    if test_samples_amt < 2:
        raise ValueError('test_samples_amt must be greater than 2')

    if iters_per_step < 1:
        raise ValueError('iters_per_step must be greater than 1')

    if max_steps < 10:
        raise ValueError('max_steps must be greater than 10')

    if first_step < 1:
        raise ValueError('first_step must be greater than 1')

    loss_types_enum_values = [loss_type.value for loss_type in list(LossType)]
    if not all(loss_type in loss_types_enum_values for loss_type in loss_types):
        raise ValueError('invalid loss_types list, implemented loss_types are: ' + str(loss_types_enum_values))


    if dims:
      if not all((dim > 0 and dim <= model.dim) for dim in dims):
          raise ValueError('invalid dims list/tuple for simulation, available dims for this Model are: ' + str([dim for dim in range(1, model.dim+1)]))
      self.dims = sorted(list(dims))
    else:
      self.dims = [dim for dim in range(1, model.dim + 1)]


    self.test_samples_amt = test_samples_amt
    self.iters_per_step = iters_per_step
    self.max_steps = max_steps
    self.first_step = first_step
    self.precision = precision
    self.augmentation_until_n = augmentation_until_n
    self.loss_types = list(loss_types)
    self.model = model
    self.report = Report(self)
    self.time_spent_test = 0
    self.verbose = verbose
    self.is_notebook = is_notebook()
    self.dims_to_compare = dims_to_compare if dims_to_compare else self.dims[-2:]

  def print_N_progress(self,n: int, max_iter: int, iter_per_step: int,datapoints_fig: plt.Figure, datapoints_plot_png_path):

    """Prints the progress of the simulation for a given n ∈ N and a given number of iterations per step (iter_per_step). The progress is printed in the terminal and a plot of the ellipsoids for this model's covariance matrix and a dataset sample with n=1024 sample points is shown.


    :param n: cardinality of the dataset
    :param max_iter: max number of iterations per n ∈ N
    :param iter_per_step: number of datasets generated in one simulation step equals iter_per_step*sqrt(augmentation_until_n)/sqrt(n), if n < augmentation_until_n, else it equals iter_per_step
    :param datapoints_fig: a plot of the ellipsoids for this model's covariance matrix and a dataset sample with n=1024 sample points


    """

    # terminal output setting
    progress_stream = '----------------------------------------------------------------------------------------------------'
    n_index = (self.model.N.index(n)+1)
    N_size = len(self.model.N)
    p = int(n_index*100/N_size)

    if self.is_notebook:
      clear_output()
    else:
      cls()

    if self.is_notebook:
      plt.close()
      if datapoints_fig :
        plt.figure(figsize=(10, 4))
        fm = plt.get_current_fig_manager()
        fm.canvas.figure = datapoints_fig
        datapoints_fig.canvas = fm.canvas

        plt.figure(datapoints_fig.number)
      plt.figure(self.report.plot_with_intersection().number)
      plt.show()



    print(' progress: ', end='')
    print(progress_stream[0:p] , end='')
    print("\033[91m {}\033[00m" .format(progress_stream[p:-1]) +' '+ str(n_index)+ '/' + str(N_size))
    print('n: ' + str(n))
    print('N = ' + str(self.model.N))
    print('max_iter = ' + str(max_iter))
    print('iter_per_step = ', iter_per_step)
    print('Model: ',self.report.model_tag)
    print('Simulator: ',self.report.sim_tag)
    print('d: ', self.report.d)
    print('bayes error rate: ', self.report.loss_bayes)
    print('Data Points Plot: ', datapoints_plot_png_path,'\n')
    for dim in self.dims:
      if self.report.loss_bayes[dim] == 0:
        print('when BR = 0, it will be infered after simulation')
        break

  def print_step_report(self, i, iter_per_step, loss_sum, iter_N):
    """
    :param self: self object of the class Simulator
    :type self: Simulator
    :param i: current iteration
    :param iter_per_step: iterations per step
    :param loss_sum: loss sum
    :param iter_N: iterations per n

    """

    print('step (',int((i+1)/iter_per_step),'): ', {d : {loss_type : loss_sum[d][loss_type]/iter_N[d][loss_type]  for loss_type in self.loss_types} for d in self.dims})

  def loss_bayes_empirical(self, d):
    """This function estimates numerically the bayes error rate for a classifier :math:`h` in dictionary :math:`H`.
      - When cardinality :math:`n` is large enough, a loss function :math:`L(h)` converges to the bayes error rate for a given dimensionality :math:`d`.
      - We can estimate the Bayes Error Rate by averaging loss :math:`L(h)` values for sufficiently large :math:`n`.
      - The best choice is to opt for averaging :math:`\hat{L}(\hat{h}^{(D)})` and :math:`\hat{L}(\hat{h}^{(D)},D')` for a Dataset :math:`D` with sufficiently large cardinality :math:`n`.

    Let :math:`\hat{h}^{(D)}` be the best empirical classifier in the dictionary :math:`H` for the dataset :math:`D` with :math:`d` features and :math:`n` samples, i.e., the classifier :math:`h` in :math:`H` with the minimum empirical error rate. Under these conditions, we can define three error rates:
      - :math:`L(\hat{h}^{(D)}) =` the theoretical error rate of the classifier :math:`\hat{h}^{(D)}`
      - :math:`\hat{L}(\hat{h}^{(D)}) =` the empirical error rate of the classifier :math:`\hat{h}^{(D)}`
      - :math:`\hat{L}(\hat{h}^{(D)},D') =` the empirical error rate of the classifier :math:`\hat{h}^{(D)}` using a test dataset :math:`D'`

    The Bayes Error (also: Bayes Risk) is defined as the minimum theoretical achievable error rate:
      - :math:`\min_{h\in H} L(h) =` the smallest theoretically achievable error rate for a classifier :math:`h` in dictionary :math:`H`.

    Parameters:
      self (Simulator): self object of the class Simulator
      d (int): dimensionality :math:`d` for which the bayes error rate is estimated

    Returns:
      float: An estimation for the smallest theoretically achievable error rate for a classifier :math:`h` in dictionary :math:`H` for the given covariance matrix :math:`\Sigma`, also called Bayes Error Rate (or Bayes Risk), represented by :math:`\min_{h\in H} L(h)`.


    """

    if 'EMPIRICAL_TRAIN' in self.loss_types and 'EMPIRICAL_TEST' in self.loss_types:
      # estimate bayes error rate using the average of the last loss values computed in the simulation for the empirical loss types
      losses = [ self.report.loss_N[d]['EMPIRICAL_TRAIN'][-1] , self.report.loss_N[d]['EMPIRICAL_TEST'][-1]]
    elif 'EMPIRICAL_TRAIN' in self.loss_types and 'THEORETICAL' in self.loss_types:
      losses = [ self.report.loss_N[d]['EMPIRICAL_TRAIN'][-1] , self.report.loss_N[d]['THEORETICAL'][-1]]
    else:
      losses = [ self.report.loss_N[d][loss_type][-1] for loss_type in self.loss_types]

    return np.mean(losses)

  def generate_dataset(self, n, i):
    """This method generates all simulation required datasets for all dimensionalities :math:`d` set to be simulated, given:
      - covariance matrix :math:`\Sigma` contained in the simulator object
      - class means :math:`\mathbf{\mu_+}` and :math:`\mathbf{\mu_+}` contained in the simulator object
      - random seed :math:`i`
      - cardinality :math:`n`

    Required Simulation Datasets:
      - :math:`D = D_{train} \cup D_{test}` : dataset
      - :math:`D_{train} = D_{(+)train} \cup D_{(-)train}` : training set
      - :math:`D_{test} = D_{(+)test} \cup D_{(-)test}` : test set
      - :math:`D_{(+)train} = (\mathbf{X}_{(+)train}, \mathbf{y}_{(+)train})` : training set for class :math:`(+)`
      - :math:`D_{(-)train} = (\mathbf{X}_{(-)train}, \mathbf{y}_{(-)train})` : training set for class :math:`(-)`
      - :math:`D_{(+)test} = (\mathbf{X}_{(+)test}, \mathbf{y}_{(+)test})` : test set for class :math:`(+)`
      - :math:`D_{(-)test} = (\mathbf{X}_{(-)test}, \mathbf{y}_{(-)test})` : test set for class :math:`(-)`

    Gaussian Distribution Parameters:
      - :math:`\mathbf{\Sigma}` : covariance matrix :math:`d \\times d` for training and test sets for both classes
      - :math:`n` : cardinality for training and test sets for both classes
      - :math:`\mathbf{\mu_+}` : mean vector for class :math:`(+)`
      - :math:`\mathbf{\mu_-}` : mean vector for class :math:`(-)`
      - :math:`\mathbf{\mu_+} = \\bigcup_{i=1}^{d} \mu_{+i}`
      - :math:`\mathbf{\mu_+} = \\bigcup_{i=1}^{d} \mu_{-i}`
      - :math:`\mu_{+i} = 1` and :math:`\mu_{-i} = -1`, for :math:`i \in [1,...,d]`
      - :math:`\mathbf{\mu_-} = [\mu_{-i}]` for :math:`i \in [1,...,d] = [\mu_{-1}, ..., \mu_{-d}] = [1,...,1]`
      - :math:`\mathbf{\mu_+} = [\mu_{+i}]` for :math:`i \in [1,...,d] = [\mu_{+1}, ..., \mu_{+d}] = [-1,...,-1]`


    Parameters:
      self (Simulator): self object of the class Simulator
      n (int): cardinality :math:`n` for training and testing sets for both classes
      i (int): random seed :math:`i`

    Returns:
      dict: A dictionary containing all required simulation datasets for all dimensionalities :math:`d` set to be simulated, represented by :math:`D_{(+)train}`, :math:`D_{(-)train}`, :math:`D_{(+)test}`, :math:`D_{(-)test}`.

    """

    half_n = int(n/2)

    #generate N/2 D-variate normal points for trainning images for each class
    sample_data_class_pos_train = multivariate_normal(self.model.mean_pos, self.model.cov).rvs(size=half_n, random_state=i)
    sample_data_class_neg_train = multivariate_normal(self.model.mean_neg, self.model.cov).rvs(size=half_n, random_state=i+1)

    #generate D-variate normal points for testing images
    self.time_spent_test = time.time()
    sample_data_class_pos_test = multivariate_normal(self.model.mean_pos, self.model.cov).rvs(size=self.test_samples_amt, random_state=i+2)
    sample_data_class_neg_test = multivariate_normal(self.model.mean_neg, self.model.cov).rvs(size=self.test_samples_amt, random_state=i+3)
    self.time_spent_test = time.time() - self.time_spent_test

    #generate target array for 2 classes: pos (class 1), neg (class 0) training images
    sample_target_class_pos_train = np.full((half_n, 1), 1, dtype=int)
    sample_target_class_neg_train = np.full((half_n, 1), 0, dtype=int)

    #generate target array for 2 classes: pos (class 1), neg (class 0) testing images
    self.time_spent_test = time.time() - self.time_spent_test
    sample_target_class_pos_test = np.full((self.test_samples_amt, 1), 1, dtype=int)
    sample_target_class_neg_test = np.full((self.test_samples_amt, 1), 0, dtype=int)
    self.time_spent_test = time.time() - self.time_spent_test

    #concatenate images for both classes into one single array for training and testing images
    sample_data_train = np.append(sample_data_class_pos_train, sample_data_class_neg_train, axis=0) if n > 2 else np.append(sample_data_class_pos_train, sample_data_class_neg_train, axis=0).reshape(2, self.model.dim)
    self.time_spent_test = time.time() - self.time_spent_test
    sample_data_test = np.append(sample_data_class_pos_test, sample_data_class_neg_test, axis=0)
    self.time_spent_test = time.time() - self.time_spent_test

    #concatenate target for both classes into one single array for training and testing images
    sample_target_train = np.append(sample_target_class_pos_train, sample_target_class_neg_train , axis=0)
    self.time_spent_test = time.time() - self.time_spent_test
    sample_target_test = np.append(sample_target_class_pos_test, sample_target_class_neg_test , axis=0)
    self.time_spent_test = time.time() - self.time_spent_test

    dataset = {'train': {'data': sample_data_train, 'target':sample_target_train},
          'test': {'data': sample_data_test, 'target':sample_target_test}}

    return dataset


  def run(self):

    """run the simulation

    :param self: simulation object
    :type self: Simulator
    :return: simulation report object
    :rtype: Report

    """

    # get the start time
    st = time.time()


    # iniciate  counters
    prev_mean_mesure = {dim : {loss_type: 0 for loss_type in self.loss_types} for dim in self.dims}

    # compute min(L(h)) and d for each dimension
    self.report.loss_bayes = {d : loss_bayes_analytical(np.array(self.model.cov[0:d]).T[0:d].T) for d in self.dims}
    self.report.d = {d : distance_from_origin_to_intersection_between_normalized_ellipsoid_and_main_diagonal(np.array(self.model.cov[0:d]).T[0:d].T) for d in self.dims}


    fig = self.model.data_points_plot
    datapoints_plot_png_path = self.model.save_data_points_plot_as_png(verbose=self.verbose)

    N = self.model.N
    N_start_size = len(self.model.N)

    ## for each cardinality n in N  do the simulation for each dimension d in dims
    while True:
      for n in N :

        # set iteration parameters for the current n
        augmentation_factor = math.sqrt(self.augmentation_until_n)/math.sqrt(n)
        iter_per_step = math.floor(self.iters_per_step*augmentation_factor) if n < self.augmentation_until_n else self.iters_per_step
        max_iter = iter_per_step*self.max_steps
        self.report.max_iter_N.append(max_iter)

        # iniciate sum(L) vars for each dimension and loss type
        loss_sum = {d : {loss_type : 0 for loss_type in self.loss_types} for d in self.dims}
        iter_N = {d : {loss_type : 0 for loss_type in self.loss_types} for d in self.dims}

        #iniciate control switches for each dimension and loss type
        switch = {'first_step': True,'train': { d : {loss_type : True if loss_type in self.loss_types else False for loss_type in [loss_type.value for loss_type in LossType ]} for d in self.dims}, 'dims': {d : True for d in self.dims} }

        # terminal output N progress bar
        if self.verbose:
          self.print_N_progress(n, max_iter,iter_per_step,fig,datapoints_plot_png_path)

        # for each iteration i in max_iter do the simulation for each dimension d in dims and estimate L(h) for each loss type
        for i in range(max_iter):

          stopwatch_dataset = time.time()
          dataset = self.generate_dataset(n,i)
          stopwatch_dataset = time.time() - stopwatch_dataset
          self.report.time_spent['total'] += stopwatch_dataset

          dims_on_by_loss_types = {loss_type : sum([switch['train'][dim][loss_type] for dim in self.dims]) for loss_type in self.loss_types}

          ## for each dimension d in dims do the simulation and estimate L(h) for each loss type
          for d in self.dims:

            if switch['dims'][d] :
              stopwatch_train = time.time()
              clf = linear_svm_train(d, dataset['train'])
              stopwatch_train = time.time() - stopwatch_train
              self.report.time_spent['total'] += stopwatch_train
              self.report.time_spent[d] += stopwatch_train
              self.report.time_spent[d] = time.time() - self.report.time_spent[d]

              if switch['train'][d][LossType.THEORETICAL.value]:
                self.report.time_spent[LossType.THEORETICAL.value] = time.time() - self.report.time_spent[LossType.THEORETICAL.value] - stopwatch_dataset/dims_on_by_loss_types[LossType.THEORETICAL.value] - stopwatch_train + self.time_spent_test/dims_on_by_loss_types[LossType.THEORETICAL.value]
                self.report.time_spent['total'] = time.time() - self.report.time_spent['total']
                loss_sum[d][LossType.THEORETICAL.value] += loss_theoretical(clf,np.array(self.model.cov[0:d]).T[0:d].T)
                self.report.time_spent[LossType.THEORETICAL.value] = time.time() - self.report.time_spent[LossType.THEORETICAL.value]
                self.report.time_spent['total'] = time.time() - self.report.time_spent['total']
                iter_N[d][LossType.THEORETICAL.value] += 1

              if switch['train'][d][LossType.EMPIRICALTRAIN.value]:
                self.report.time_spent[LossType.EMPIRICALTRAIN.value] = time.time() - self.report.time_spent[LossType.EMPIRICALTRAIN.value] - stopwatch_dataset/dims_on_by_loss_types[LossType.EMPIRICALTRAIN.value] - stopwatch_train + self.time_spent_test/dims_on_by_loss_types[LossType.EMPIRICALTRAIN.value]
                self.report.time_spent['total'] = time.time() - self.report.time_spent['total']
                loss_sum[d][LossType.EMPIRICALTRAIN.value] += loss_empirical(clf, dataset['train'])
                self.report.time_spent[LossType.EMPIRICALTRAIN.value] = time.time() - self.report.time_spent[LossType.EMPIRICALTRAIN.value]
                self.report.time_spent['total'] = time.time() - self.report.time_spent['total']
                iter_N[d][LossType.EMPIRICALTRAIN.value] += 1

              if switch['train'][d][LossType.EMPIRICALTEST.value]:
                self.report.time_spent[LossType.EMPIRICALTEST.value] = time.time() - self.report.time_spent[LossType.EMPIRICALTEST.value] - stopwatch_dataset/dims_on_by_loss_types[LossType.EMPIRICALTEST.value] - stopwatch_train
                self.report.time_spent['total'] = time.time() - self.report.time_spent['total']
                loss_sum[d][LossType.EMPIRICALTEST.value] += loss_empirical(clf, dataset['test'])
                self.report.time_spent[LossType.EMPIRICALTEST.value] = time.time() - self.report.time_spent[LossType.EMPIRICALTEST.value]
                self.report.time_spent['total'] = time.time() - self.report.time_spent['total']
                iter_N[d][LossType.EMPIRICALTEST.value] += 1

              self.report.time_spent[d] = time.time() - self.report.time_spent[d]

          ## simulate for at least first_step*iter_per_step iterations before checking for convergence
          if (i+1) % iter_per_step == 0:
            if switch['first_step']:

              ## start checking for convergence after turn off first_step switch
              if (i+1) == self.first_step*iter_per_step:
                switch['first_step'] = False
                for d in self.dims:
                  for loss_type in self.loss_types:
                    prev_mean_mesure[d][loss_type] = loss_sum[d][loss_type]/iter_N[d][loss_type]
            ## check for convergence and turn off switch if necessary for each loss type and dimension d
            else:
              for d in self.dims:
                for loss_type in self.loss_types:
                  if switch['train'][d][loss_type]:
                    if abs(loss_sum[d][loss_type]/iter_N[d][loss_type] - prev_mean_mesure[d][loss_type]) < (self.precision if loss_type != LossType.EMPIRICALTRAIN.value else self.precision*1e-2) :
                      switch['train'][d][loss_type] = False
                      if self.verbose:
                        print('SWITCH OFF: (', loss_type, ',', d,')')
                      if (np.unique(list(switch['train'][d].values())).tolist()) == [False] :
                        switch['dims'][d] = False
                        if self.verbose:
                          print('SWITCH OFF: dim (' , d ,')')
                    else:
                      prev_mean_mesure[d][loss_type] = loss_sum[d][loss_type]/iter_N[d][loss_type]

            ## print report for each step
            if self.verbose:
              self.print_step_report(i,iter_per_step,loss_sum,iter_N)

            ## check if all dimensions are turned off and stop simulation if so
            if np.unique(list(switch['dims'].values())).tolist() == [False]:
              break

        ## compute time spent for each cardinality n
        try:
          self.report.time_spent['n'][int(math.log(n,2)-1)] = self.report.time_spent['total'] - sum(self.report.time_spent['n'][:int(math.log(n,2)-1)])
        except IndexError:
          self.report.time_spent['n'].append(self.report.time_spent['total'] - sum(self.report.time_spent['n']))

        ## report loss and iter number for each dim and loss type
        for d in self.dims:
          for loss_type in self.loss_types:
            self.report.loss_N[d][loss_type].append(loss_sum[d][loss_type]/iter_N[d][loss_type])
            self.report.iter_N[d][loss_type].append(iter_N[d][loss_type])

      ## after simulation for each cardinality n in self.dims is finished, some tests are performed to check if simulation should be continued
      finish = True

      if 'EMPIRICAL_TEST' in self.loss_types:

        ## test if intersection point between dim d and dim d-1 is found and continue simulation for one more cardinality n if not
        if self.report.loss_N[self.dims[-1]]['EMPIRICAL_TEST'][-1] > self.report.loss_N[self.dims[-2]]['EMPIRICAL_TEST'][-1]:
          finish = False

        ## test if Loss Functions are not yet converged and continue simulation for one more cardinality n if not in order to estimate the Bayes risk
        if self.report.loss_bayes[self.model.dim] == 0 and abs(self.report.loss_N[self.model.dim]['EMPIRICAL_TEST'][-1] - self.report.loss_N[self.model.dim]['EMPIRICAL_TEST'][-2]) > 0.001:
          finish = False
        if 'EMPIRICAL_TRAIN' in self.loss_types:
          if self.report.loss_bayes[self.model.dim] == 0 and abs(self.report.loss_N[self.model.dim]['EMPIRICAL_TEST'][-1] - self.report.loss_N[self.model.dim]['EMPIRICAL_TRAIN'][-1]) > 0.001:
            finish = False


      if 'THEORETICAL' in self.loss_types:

        ## test if intersection point between dim d and dim d-1 is found and continue simulation for one more cardinality n if not
        if self.report.loss_N[self.dims[-1]]['THEORETICAL'][-1] > self.report.loss_N[self.dims[-2]]['THEORETICAL'][-1]:
          finish = False

        ## test if Loss Functions are not yet converged and continue simulation for one more cardinality n if not in order to estimate the Bayes risk
        if self.report.loss_bayes[self.model.dim] == 0 and abs(self.report.loss_N[self.model.dim]['THEORETICAL'][-1] - self.report.loss_N[self.model.dim]['THEORETICAL'][-2]) > 0.001:
          finish = False
        if 'EMPIRICAL_TRAIN' in self.loss_types:
          if self.report.loss_bayes[self.model.dim] == 0 and abs(self.report.loss_N[self.model.dim]['THEORETICAL'][-1] - self.report.loss_N[self.model.dim]['EMPIRICAL_TRAIN'][-1]) > 0.001:
            finish = False

      ## stop simulation if all Loss Functions are converged and intersection points are found, or if max cardinality model.max_n is reached
      if finish == True or max(N) >= self.model.max_n:
        break
      else:
        if self.model.N[-1]/self.model.N[-2] == 2:
          new_N = 2*max(N)
          self.model.N.append(new_N)
          N = [new_N]
        else:
          new_N = 2 + max(N)
          self.model.N.append(new_N)
          N = [new_N]

    ## estimate Bayes risk for each dimension if not yet estimated by analytical method
    for d in list(self.report.loss_bayes.keys()):
      if not self.report.loss_bayes[d]:
        self.report.loss_bayes[d] = self.loss_bayes_empirical(d)

    plt.close()
    ## set and save the final plot

    self.report.loss_plot = self.report.plot_with_intersection()

    ## print the final report
    self.report.print_report()

    ## save the final plot
    self.report.save_loss_plot_as_png(verbose=self.verbose)
    plt.close()

    # get the end time
    et = time.time()

    # get the execution time
    elapsed_time = et - st

    if self.verbose:
      print('Execution time:', elapsed_time/3600, 'h\n')

    ## transform time spent from seconds to hours
    self.report.duration = elapsed_time/3600
    for key in list(self.report.time_spent.keys()):
      if not isinstance(self.report.time_spent[key], list):
        self.report.time_spent[key] = self.report.time_spent[key]/3600
      else:
        self.report.time_spent[key] = [self.report.time_spent[key][i]/3600 for i in range(len(self.report.time_spent[key]))]






from setuptools import setup, find_packages

setup(
    name='slacgs',
    version='0.1.0',
    packages=find_packages(),
    install_requires=[
        # List any dependencies your package requires
        'numpy>=1.19.5',
        'matplotlib>=3.3.3',
        'shapely>=2.0.1',
        'ipython>=7.34.0',
        'scikit-learn>=1.2.2',
        'scipy>=1.10.1',
        'pygsheets>=2.0.6',
        'google-api-python-client>=2.83.0',
        'cryptography>=41.0.2',
        'Pillow>=8.1.0',
        'pandas>=1.5.3',
        'setuptools>=51.3.3',
        'tabulate>=0.9.0',
        'reportlab>=4.0.4'

    ],
    author='Paulo Azevedo',
    author_email='paulorenatoaz@dcc.ufrj.br',
    description='A Simulator for Loss Analysis of Classifiers on Gaussian Samples.',
    url='https://github.com/paulorenatoaz/slacgs',
)
